import dataclasses
import datetime
import platform
import re
import sys
import typing
from functools import lru_cache
from typing import Any, Callable, Dict, Iterator, Optional, Sequence, Union

import pydantic
from requests import HTTPError

from weave import version
from weave.legacy.weave import ref_base, urls
from weave.trace import call_context, trace_sentry
from weave.trace.async_job_queue import AsyncJobQueue
from weave.trace.client_context import weave_client as weave_client_context
from weave.trace.exception import exception_to_json_str
from weave.trace.feedback import FeedbackQuery, RefFeedbackQuery
from weave.trace.object_record import (
    ObjectRecord,
    dataclass_object_record,
    pydantic_object_record,
)
from weave.trace.op import Op, is_op, maybe_unbind_method
from weave.trace.op import op as op_deco
from weave.trace.refs import CallRef, ObjectRef, OpRef, Ref, TableRef, parse_op_uri
from weave.trace.serialize import from_json, isinstance_namedtuple, to_json
from weave.trace.table import Table
from weave.trace.util import deprecated
from weave.trace.vals import WeaveObject, WeaveTable, make_trace_obj
from weave.trace_server.ids import generate_id
from weave.trace_server.trace_server_interface import (
    CallEndReq,
    CallSchema,
    CallsDeleteReq,
    CallsFilter,
    CallsQueryReq,
    CallStartReq,
    CallUpdateReq,
    CostCreateInput,
    CostCreateReq,
    CostCreateRes,
    CostPurgeReq,
    CostQueryOutput,
    CostQueryReq,
    EndedCallSchemaForInsert,
    ObjCreateReq,
    ObjectVersionFilter,
    ObjQueryReq,
    ObjReadReq,
    ObjSchema,
    ObjSchemaForInsert,
    Query,
    RefsReadBatchReq,
    StartedCallSchemaForInsert,
    TableCreateReq,
    TableSchemaForInsert,
    TraceServerInterface,
)
from weave.trace_server_bindings.remote_http_trace_server import RemoteHTTPTraceServer

# Controls if objects can have refs to projects not the WeaveClient project.
# If False, object refs with with mismatching projects will be recreated.
# If True, use existing ref to object in other project.
ALLOW_MIXED_PROJECT_REFS = False


def dataclasses_asdict_one_level(obj: Any) -> typing.Dict[str, Any]:
    # dataclasses.asdict is recursive. We don't want that when json encoding
    return {f.name: getattr(obj, f.name) for f in dataclasses.fields(obj)}


# TODO: unused


def get_obj_name(val: Any) -> str:
    name = getattr(val, "name", None)
    if name is None:
        if isinstance(val, ObjectRecord):
            name = val._class_name
        else:
            name = f"{val.__class__.__name__}"
    if not isinstance(name, str):
        raise ValueError(f"Object's name attribute is not a string: {name}")
    return name


def get_ref(obj: Any) -> Optional[ObjectRef]:
    return getattr(obj, "ref", None)


def remove_ref(obj: Any) -> None:
    if get_ref(obj) is not None:
        if "ref" in obj.__dict__:  # for methods
            obj.__dict__["ref"] = None
        else:
            obj.ref = None


def set_ref(obj: Any, ref: Optional[Ref]) -> None:
    """Try to set the ref on "any" object.

    We use increasingly complex methods to try to set the ref
    to support different kinds of objects. This will still
    fail for python primitives, but those can't be traced anyway.
    """
    try:
        obj.ref = ref
    except:
        try:
            setattr(obj, "ref", ref)
        except:
            try:
                obj.__dict__["ref"] = ref
            except:
                raise ValueError(f"Failed to set ref on object of type {type(obj)}")


def _get_direct_ref(obj: Any) -> Optional[Ref]:
    if isinstance(obj, WeaveTable):
        # TODO: this path is odd. We want to use table_ref when serializing
        # which is the direct ref to the table. But .ref on WeaveTable is
        # the "container ref", ie a ref to the root object that the WeaveTable
        # is within, with extra pointing to the table.
        return obj.table_ref
    return getattr(obj, "ref", None)


def map_to_refs(obj: Any) -> Any:
    if isinstance(obj, Ref):
        return obj
    if ref := _get_direct_ref(obj):
        return ref

    if isinstance(obj, ObjectRecord):
        return obj.map_values(map_to_refs)
    elif isinstance(obj, (pydantic.BaseModel, pydantic.v1.BaseModel)):
        obj_record = pydantic_object_record(obj)
        return obj_record.map_values(map_to_refs)
    elif dataclasses.is_dataclass(obj):
        obj_record = dataclass_object_record(obj)
        return obj_record.map_values(map_to_refs)
    elif isinstance(obj, Table):
        return obj.ref
    elif isinstance(obj, WeaveTable):
        return obj.ref
    elif isinstance(obj, list):
        return [map_to_refs(v) for v in obj]
    elif isinstance(obj, dict):
        return {k: map_to_refs(v) for k, v in obj.items()}

    # This path should only be reached if the object is both:
    # 1. A `WeaveObject`; and
    # 2. Has been dirtied (edited in any way), causing obj.ref=None
    elif isinstance(obj, WeaveObject):
        return map_to_refs(obj._val)

    return obj


@dataclasses.dataclass
class Call:
    """A Call represents a single operation that was executed as part of a trace."""

    op_name: str
    trace_id: str
    project_id: str
    parent_id: Optional[str]
    inputs: dict
    id: Optional[str] = None
    output: Any = None
    exception: Optional[str] = None
    summary: Optional[dict] = None
    display_name: Optional[Union[str, Callable[["Call"], str]]] = None
    attributes: Optional[dict] = None
    started_at: Optional[datetime.datetime] = None
    ended_at: Optional[datetime.datetime] = None
    deleted_at: Optional[datetime.datetime] = None
    # These are the live children during logging
    _children: list["Call"] = dataclasses.field(default_factory=list)

    _feedback: Optional[RefFeedbackQuery] = None

    @property
    def func_name(self) -> str:
        """
        The decorated function's name that produced this call.

        This is different from `op_name` which is usually the ref of the op.
        """
        if self.op_name.startswith("weave:///"):
            ref = parse_op_uri(self.op_name)
            return ref.name

        return self.op_name

    @property
    def feedback(self) -> RefFeedbackQuery:
        if not self.id:
            raise ValueError("Can't get feedback for call without ID")
        if self._feedback is None:
            project_parts = self.project_id.split("/")
            if len(project_parts) != 2:
                raise ValueError(f"Invalid project_id: {self.project_id}")
            entity, project = project_parts
            weave_ref = CallRef(entity, project, self.id)
            self._feedback = RefFeedbackQuery(weave_ref.uri())
        return self._feedback

    @property
    def ui_url(self) -> str:
        project_parts = self.project_id.split("/")
        if len(project_parts) != 2:
            raise ValueError(f"Invalid project_id: {self.project_id}")
        entity, project = project_parts
        if not self.id:
            raise ValueError("Can't get URL for call without ID")
        return urls.redirect_call(entity, project, self.id)

    # These are the children if we're using Call at read-time
    def children(self) -> "CallsIter":
        client = weave_client_context.require_weave_client()
        if not self.id:
            raise ValueError("Can't get children of call without ID")
        return CallsIter(
            client.server,
            self.project_id,
            CallsFilter(parent_ids=[self.id]),
        )

    def delete(self) -> bool:
        """Delete the call."""
        client = weave_client_context.require_weave_client()
        return client.delete_call(call=self)

    def set_display_name(self, name: Optional[str]) -> None:
        """
        Set the display name for the call.

        Args:
            name: The display name to set for the call.

        Example:

        ```python
        result, call = my_function.call("World")
        call.set_display_name("My Custom Display Name")
        ```
        """
        if name == "":
            raise ValueError(
                "Display name cannot be empty. To remove the display_name, set name=None or use remove_display_name."
            )
        if name == self.display_name:
            return
        client = weave_client_context.require_weave_client()
        client._set_call_display_name(call=self, display_name=name)
        self.display_name = name

    def remove_display_name(self) -> None:
        self.set_display_name(None)


class CallsIter:
    server: TraceServerInterface
    filter: CallsFilter
    include_costs: bool

    def __init__(
        self,
        server: TraceServerInterface,
        project_id: str,
        filter: CallsFilter,
        include_costs: bool = False,
    ) -> None:
        self.server = server
        self.project_id = project_id
        self.filter = filter
        self._page_size = 1000
        self.include_costs = include_costs

    # seems like this caching should be on the server, but it's here for now...
    @lru_cache
    def _fetch_page(self, index: int) -> list[CallSchema]:
        # caching in here means that any other CallsIter objects would also
        # benefit from the cache
        response = self.server.calls_query(
            CallsQueryReq(
                project_id=self.project_id,
                filter=self.filter,
                offset=index * self._page_size,
                limit=self._page_size,
                include_costs=self.include_costs,
            )
        )
        return response.calls

    def _get_one(self, index: int) -> WeaveObject:
        if index < 0:
            raise IndexError("Negative indexing not supported")

        page_index = index // self._page_size
        page_offset = index % self._page_size

        calls = self._fetch_page(page_index)
        if page_offset >= len(calls):
            raise IndexError(f"Index {index} out of range")

        call = calls[page_offset]
        entity, project = self.project_id.split("/")
        return make_client_call(entity, project, call, self.server)

    def _get_slice(self, key: slice) -> Iterator[WeaveObject]:
        if (start := key.start or 0) < 0:
            raise ValueError("Negative start not supported")
        if (stop := key.stop) is not None and stop < 0:
            raise ValueError("Negative stop not supported")
        if (step := key.step or 1) < 0:
            raise ValueError("Negative step not supported")

        i = start
        while stop is None or i < stop:
            try:
                yield self._get_one(i)
            except IndexError:
                break
            i += step

    def __getitem__(
        self, key: Union[slice, int]
    ) -> Union[WeaveObject, list[WeaveObject]]:
        if isinstance(key, slice):
            return list(self._get_slice(key))
        return self._get_one(key)

    def __iter__(self) -> typing.Iterator[WeaveObject]:
        return self._get_slice(slice(0, None, 1))


def make_client_call(
    entity: str, project: str, server_call: CallSchema, server: TraceServerInterface
) -> WeaveObject:
    output = server_call.output
    call = Call(
        op_name=server_call.op_name,
        project_id=server_call.project_id,
        trace_id=server_call.trace_id,
        parent_id=server_call.parent_id,
        id=server_call.id,
        inputs=from_json(server_call.inputs, server_call.project_id, server),
        output=from_json(output, server_call.project_id, server),
        summary=dict(server_call.summary) if server_call.summary is not None else None,
        display_name=server_call.display_name,
        attributes=server_call.attributes,
        started_at=server_call.started_at,
        ended_at=server_call.ended_at,
        deleted_at=server_call.deleted_at,
    )
    if call.id is None:
        raise ValueError("Call ID is None")
    return WeaveObject(call, CallRef(entity, project, call.id), server, None)


def sum_dict_leaves(dicts: list[dict]) -> dict:
    # dicts is a list of dictionaries, that may or may not
    # have nested dictionaries. Sum all the leaves that match
    result: dict = {}
    for d in dicts:
        for k, v in d.items():
            if isinstance(v, dict):
                result[k] = sum_dict_leaves([result.get(k, {}), v])
            elif v is not None:
                result[k] = result.get(k, 0) + v
    return result


class WeaveKeyDict(dict):
    """A dict representing the 'weave' subdictionary of a call's attributes.

    This dictionary is not intended to be set directly.
    """

    def __setitem__(self, key: Any, value: Any) -> None:
        raise KeyError("Cannot modify `weave` dict directly -- for internal use only!")


class AttributesDict(dict):
    """A dict representing the attributes of a call.

    The `weave` key is reserved for internal use and cannot be set directly.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        dict.__setitem__(self, "weave", WeaveKeyDict())

        if kwargs:
            for key, value in kwargs.items():
                if key == "weave":
                    if isinstance(value, dict):
                        for subkey, subvalue in value.items():
                            self._set_weave_item(subkey, subvalue)
                else:
                    self[key] = value

    def __setitem__(self, key: Any, value: Any) -> None:
        if key == "weave":
            raise KeyError("Cannot set 'weave' directly -- for internal use only!")
        super().__setitem__(key, value)

    def _set_weave_item(self, subkey: Any, value: Any) -> None:
        """Internal method to set items in the 'weave' subdictionary."""
        dict.__setitem__(self["weave"], subkey, value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"


class WeaveClient:
    server: TraceServerInterface
    async_job_queue: AsyncJobQueue

    """
    A client for interacting with the Weave trace server.

    Args:
        entity: The entity name.
        project: The project name.
        server: The server to use for communication.
        ensure_project_exists: Whether to ensure the project exists on the server.
    """

    def __init__(
        self,
        entity: str,
        project: str,
        server: TraceServerInterface,
        ensure_project_exists: bool = True,
    ):
        self.entity = entity
        self.project = project
        self.server = server
        self._anonymous_ops: dict[str, Op] = {}
        self.async_job_queue = AsyncJobQueue()
        self.ensure_project_exists = ensure_project_exists

        if ensure_project_exists:
            resp = self.server.ensure_project_exists(entity, project)
            # Set Client project name with updated project name
            self.project = resp.project_name

        self._server_is_flushable = False
        if isinstance(self.server, RemoteHTTPTraceServer):
            self._server_is_flushable = self.server.should_batch

    ################ High Level Convenience Methods ################

    @trace_sentry.global_trace_sentry.watch()
    def save(self, val: Any, name: str, branch: str = "latest") -> Any:
        """Do not call directly, use weave.publish() instead.

        Args:
            val: The object to save.
            name: The name to save the object under.
            branch: The branch to save the object under. Defaults to "latest".

        Returns:
            A deserialized version of the saved object.
        """
        # Adding a second comment line for developers that is not a docstring:
        # Save an object to the weave server and return a deserialized version of it.

        # Note: This is sort of a weird method becuase:
        # 1. It returns a deserialized version of the object (which will often not pass type-checks)
        # 2. It is slow (because it re-downloads the object from the weave server)
        # 3. It explicitly filters out non ObjectRefs, which seems like a useless constraint.
        #
        # Because of these reasons, `weave.publish()` directly calls `_save_object()`
        # and then returns the raw ObjectRef. I (Tim) think we should consider refactoring
        # `save()`. I am not sure when as an end user you would ever want to use this method.

        ref = self._save_object(val, name, branch)
        if not isinstance(ref, ObjectRef):
            raise ValueError(f"Expected ObjectRef, got {ref}")
        return self.get(ref)

    @trace_sentry.global_trace_sentry.watch()
    def get(self, ref: ObjectRef) -> Any:
        project_id = f"{ref.entity}/{ref.project}"
        try:
            read_res = self.server.obj_read(
                ObjReadReq(
                    project_id=project_id,
                    object_id=ref.name,
                    digest=ref.digest,
                )
            )
        except HTTPError as e:
            if e.response is not None and e.response.status_code == 404:
                raise ValueError(f"Unable to find object for ref uri: {ref.uri()}")
            raise

        # At this point, `ref.digest` is one of three things:
        # 1. "latest" - the user asked for the latest version of the object
        # 2. "v###" - the user asked for a specific version of the object
        # 3. The actual digest.
        #
        # However, we always want to resolve the ref to the digest. So
        # here, we just directly assign the digest.
        ref = dataclasses.replace(ref, digest=read_res.obj.digest)

        data = read_res.obj.val

        # If there is a ref-extra, we should resolve it. Rather than walking
        # the object, it is more efficient to directly query for the data and
        # let the server resolve it.
        if ref.extra:
            try:
                ref_read_res = self.server.refs_read_batch(
                    RefsReadBatchReq(refs=[ref.uri()])
                )
            except HTTPError as e:
                if e.response is not None and e.response.status_code == 404:
                    raise ValueError(f"Unable to find object for ref uri: {ref.uri()}")
                raise
            if not ref_read_res.vals:
                raise ValueError(f"Unable to find object for ref uri: {ref.uri()}")
            data = ref_read_res.vals[0]

        val = from_json(data, project_id, self.server)

        return make_trace_obj(val, ref, self.server, None)

    ################ Query API ################

    @trace_sentry.global_trace_sentry.watch()
    def get_calls(
        self,
        filter: Optional[CallsFilter] = None,
        include_costs: Optional[bool] = False,
    ) -> CallsIter:
        if filter is None:
            filter = CallsFilter()

        return CallsIter(
            self.server, self._project_id(), filter, include_costs or False
        )

    @deprecated(new_name="get_calls")
    def calls(
        self,
        filter: Optional[CallsFilter] = None,
        include_costs: Optional[bool] = False,
    ) -> CallsIter:
        return self.get_calls(filter=filter, include_costs=include_costs)

    @trace_sentry.global_trace_sentry.watch()
    def get_call(
        self, call_id: str, include_costs: Optional[bool] = False
    ) -> WeaveObject:
        response = self.server.calls_query(
            CallsQueryReq(
                project_id=self._project_id(),
                filter=CallsFilter(call_ids=[call_id]),
                include_costs=include_costs,
            )
        )
        if not response.calls:
            raise ValueError(f"Call not found: {call_id}")
        response_call = response.calls[0]
        return make_client_call(self.entity, self.project, response_call, self.server)

    @deprecated(new_name="get_call")
    def call(self, call_id: str, include_costs: Optional[bool] = False) -> WeaveObject:
        return self.get_call(call_id=call_id, include_costs=include_costs)

    @trace_sentry.global_trace_sentry.watch()
    def create_call(
        self,
        op: Union[str, Op],
        inputs: dict,
        parent: Optional[Call] = None,
        attributes: Optional[dict] = None,
        display_name: Optional[Union[str, Callable[[Call], str]]] = None,
        *,
        use_stack: bool = True,
    ) -> Call:
        """Create, log, and push a call onto the runtime stack.

        Args:
            op: The operation producing the call, or the name of an anonymous operation.
            inputs: The inputs to the operation.
            parent: The parent call. If parent is not provided, the current run is used as the parent.
            display_name: The display name for the call. Defaults to None.
            attributes: The attributes for the call. Defaults to None.
            use_stack: Whether to push the call onto the runtime stack. Defaults to True.

        Returns:
            The created Call object.
        """
        if isinstance(op, str):
            if op not in self._anonymous_ops:
                self._anonymous_ops[op] = _build_anonymous_op(op)
            op = self._anonymous_ops[op]
        if is_op(op):
            unbound_op = maybe_unbind_method(op)
            op_def_ref = self._save_op(unbound_op)
            op_str = op_def_ref.uri()
        else:
            op_str = op

        inputs_redacted = redact_sensitive_keys(inputs)
        if op.postprocess_inputs:
            inputs_postprocessed = op.postprocess_inputs(inputs_redacted)
        else:
            inputs_postprocessed = inputs_redacted

        self._save_nested_objects(inputs_postprocessed)
        inputs_with_refs = map_to_refs(inputs_postprocessed)
        call_id = generate_id()

        if parent is None and use_stack:
            parent = call_context.get_current_call()

        if parent:
            trace_id = parent.trace_id
            parent_id = parent.id
        else:
            trace_id = generate_id()
            parent_id = None

        if attributes is None:
            attributes = {}

        attributes = AttributesDict(**attributes)
        attributes._set_weave_item("client_version", version.VERSION)
        attributes._set_weave_item("source", "python-sdk")
        attributes._set_weave_item("os_name", platform.system())
        attributes._set_weave_item("os_version", platform.version())
        attributes._set_weave_item("os_release", platform.release())
        attributes._set_weave_item("sys_version", sys.version)

        call = Call(
            op_name=op_str,
            project_id=self._project_id(),
            trace_id=trace_id,
            parent_id=parent_id,
            id=call_id,
            inputs=inputs_with_refs,
            attributes=attributes,
        )
        # feels like this should be in post init, but keping here
        # because the func needs to be resolved for schema insert below
        if callable(name_func := display_name):
            display_name = name_func(call)
        call.display_name = display_name
        if parent is not None:
            parent._children.append(call)

        current_wb_run_id = safe_current_wb_run_id()
        check_wandb_run_matches(current_wb_run_id, self.entity, self.project)

        started_at = datetime.datetime.now(tz=datetime.timezone.utc)
        project_id = self._project_id()

        self.async_job_queue.submit_job(
            send_start_call,
            project_id=project_id,
            call_id=call_id,
            op_str=op_str,
            trace_id=trace_id,
            started_at=started_at,
            display_name=display_name,
            parent_id=parent_id,
            inputs_with_refs=inputs_with_refs,
            attributes=attributes,
            current_wb_run_id=current_wb_run_id,
            server=self.server,
        )

        if use_stack:
            call_context.push_call(call)

        return call

    @trace_sentry.global_trace_sentry.watch()
    def finish_call(
        self,
        call: Call,
        output: Any = None,
        exception: Optional[BaseException] = None,
        *,
        postprocess_output: Optional[Callable[..., Any]] = None,
    ) -> None:
        original_output = output

        if postprocess_output:
            postprocessed_output = postprocess_output(original_output)
        else:
            postprocessed_output = original_output
        self._save_nested_objects(postprocessed_output)

        output = map_to_refs(postprocessed_output)
        call.output = output

        # Summary handling
        summary = {}
        if call._children:
            summary = sum_dict_leaves([child.summary or {} for child in call._children])
        elif (
            isinstance(original_output, dict)
            and "usage" in original_output
            and "model" in original_output
        ):
            summary["usage"] = {}
            summary["usage"][original_output["model"]] = {
                "requests": 1,
                **original_output["usage"],
            }
        elif hasattr(original_output, "usage") and hasattr(original_output, "model"):
            # Handle the cases where we are emitting an object instead of a pre-serialized dict
            # In fact, this is going to become the more common case
            model = original_output.model
            usage = original_output.usage
            if isinstance(usage, pydantic.BaseModel):
                usage = usage.model_dump(exclude_unset=True)
            if isinstance(usage, dict) and isinstance(model, str):
                summary["usage"] = {}
                summary["usage"][model] = {"requests": 1, **usage}

        # Exception Handling
        exception_str: Optional[str] = None
        if exception:
            exception_str = exception_to_json_str(exception)
            call.exception = exception_str

        project_id = self._project_id()
        ended_at = datetime.datetime.now(tz=datetime.timezone.utc)
        self.async_job_queue.submit_job(
            send_end_call,
            project_id=project_id,
            call_id=call.id,
            ended_at=ended_at,
            output=output,
            summary=summary,
            exception_str=exception_str,
            server=self.server,
        )

        # Descendent error tracking disabled til we fix UI
        # Add this call's summary after logging the call, so that only
        # descendents are included in what we log
        # summary.setdefault("descendants", {}).setdefault(
        #     call.op_name, {"successes": 0, "errors": 0}
        # )["successes"] += 1
        call.summary = summary
        call_context.pop_call(call.id)

    @trace_sentry.global_trace_sentry.watch()
    def fail_call(self, call: Call, exception: BaseException) -> None:
        """Fail a call with an exception. This is a convenience method for finish_call."""
        return self.finish_call(call, exception=exception)

    @trace_sentry.global_trace_sentry.watch()
    def delete_call(self, call: Call) -> None:
        self.server.calls_delete(
            CallsDeleteReq(
                project_id=self._project_id(),
                call_ids=[call.id],
            )
        )

    def get_feedback(
        self,
        query: Optional[Union[Query, str]] = None,
        *,
        reaction: Optional[str] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> FeedbackQuery:
        """Query project for feedback.

        Examples:
            ```python
            # Fetch a specific feedback object.
            # Note that this still returns a collection, which is expected
            # to contain zero or one item(s).
            client.get_feedback("1B4082A3-4EDA-4BEB-BFEB-2D16ED59AA07")

            # Find all feedback objects with a specific reaction.
            client.get_feedback(reaction="👍", limit=10)
            ```

        Args:
            query: A mongo-style query expression. For convenience, also accepts a feedback UUID string.
            reaction: For convenience, filter by a particular reaction emoji.
            offset: The offset to start fetching feedback objects from.
            limit: The maximum number of feedback objects to fetch.

        Returns:
            A FeedbackQuery object.
        """
        expr: dict[str, Any] = {
            "$eq": [
                {"$literal": "1"},
                {"$literal": "1"},
            ],
        }
        if isinstance(query, str):
            expr = {
                "$eq": [
                    {"$getField": "id"},
                    {"$literal": query},
                ],
            }
        elif isinstance(query, Query):
            expr = query.expr_.dict()

        if reaction:
            expr = {
                "$and": [
                    expr,
                    {
                        "$eq": [
                            {"$getField": "feedback_type"},
                            {"$literal": "wandb.reaction.1"},
                        ],
                    },
                    {
                        "$eq": [
                            {"$getField": "payload.emoji"},
                            {"$literal": reaction},
                        ],
                    },
                ]
            }
        rewritten_query = Query(**{"$expr": expr})

        return FeedbackQuery(
            entity=self.entity,
            project=self.project,
            query=rewritten_query,
            offset=offset,
            limit=limit,
            show_refs=True,
        )

    @deprecated(new_name="get_feedback")
    def feedback(
        self,
        query: Optional[Union[Query, str]] = None,
        *,
        reaction: Optional[str] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> FeedbackQuery:
        return self.get_feedback(
            query=query, reaction=reaction, offset=offset, limit=limit
        )

    def add_cost(
        self,
        llm_id: str,
        prompt_token_cost: float,
        completion_token_cost: float,
        effective_date: typing.Optional[datetime.datetime] = datetime.datetime.now(
            datetime.timezone.utc
        ),
        prompt_token_cost_unit: typing.Optional[str] = "USD",
        completion_token_cost_unit: typing.Optional[str] = "USD",
        provider_id: typing.Optional[str] = "default",
    ) -> CostCreateRes:
        """Add a cost to the current project.

        Examples:

            ```python
            client.add_cost(llm_id="my_expensive_custom_model", prompt_token_cost=1, completion_token_cost=2)
            client.add_cost(llm_id="my_expensive_custom_model", prompt_token_cost=500, completion_token_cost=1000, effective_date=datetime(1998, 10, 3))
            ```

        Args:
            llm_id: The ID of the LLM. eg "gpt-4o-mini-2024-07-18"
            prompt_token_cost: The cost of the prompt tokens. eg .0005
            completion_token_cost: The cost of the completion tokens. eg .0015
            effective_date: Defaults to the current date. A datetime.datetime object.
            provider_id: The provider of the LLM. Defaults to "default". eg "openai"
            prompt_token_cost_unit: The unit of the cost for the prompt tokens. Defaults to "USD". (Currently unused, will be used in the future to specify the currency type for the cost eg "tokens" or "time")
            completion_token_cost_unit: The unit of the cost for the completion tokens. Defaults to "USD". (Currently unused, will be used in the future to specify the currency type for the cost eg "tokens" or "time")

        Returns:
            A CostCreateRes object.
            Which has one field called a list of tuples called ids.
            Each tuple contains the llm_id and the id of the created cost object.
        """
        cost = CostCreateInput(
            prompt_token_cost=prompt_token_cost,
            completion_token_cost=completion_token_cost,
            effective_date=effective_date,
            prompt_token_cost_unit=prompt_token_cost_unit,
            completion_token_cost_unit=completion_token_cost_unit,
            provider_id=provider_id,
        )
        return self.server.cost_create(
            CostCreateReq(project_id=self._project_id(), costs={llm_id: cost})
        )

    def purge_costs(self, ids: Union[list[str], str]) -> None:
        """Purge costs from the current project.

        Examples:

            ```python
            client.purge_costs([ids])
            client.purge_costs(ids)
            ```

        Args:
            ids: The cost IDs to purge. Can be a single ID or a list of IDs.
        """
        if isinstance(ids, str):
            ids = [ids]
        expr = {
            "$in": [
                {"$getField": "id"},
                [{"$literal": id} for id in ids],
            ]
        }
        self.server.cost_purge(
            CostPurgeReq(project_id=self._project_id(), query=Query(**{"$expr": expr}))
        )

    def query_costs(
        self,
        query: Optional[Union[Query, str]] = None,
        llm_ids: Optional[list[str]] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> list[CostQueryOutput]:
        """Query project for costs.

        Examples:

            ```python
            # Fetch a specific cost object.
            # Note that this still returns a collection, which is expected
            # to contain zero or one item(s).
            client.query_costs("1B4082A3-4EDA-4BEB-BFEB-2D16ED59AA07")

            # Find all cost objects with a specific reaction.
            client.query_costs(llm_ids=["gpt-4o-mini-2024-07-18"], limit=10)
            ```

        Args:
            query: A mongo-style query expression. For convenience, also accepts a cost UUID string.
            llm_ids: For convenience, filter for a set of llm_ids.
            offset: The offset to start fetching cost objects from.
            limit: The maximum number of cost objects to fetch.

        Returns:
            A CostQuery object.
        """
        expr: dict[str, Any] = {
            "$eq": [
                {"$literal": "1"},
                {"$literal": "1"},
            ],
        }
        if isinstance(query, str):
            expr = {
                "$eq": [
                    {"$getField": "id"},
                    {"$literal": query},
                ],
            }
        elif isinstance(query, Query):
            expr = query.expr_

        if llm_ids:
            expr = {
                "$and": [
                    expr,
                    {
                        "$in": [
                            {"$getField": "llm_id"},
                            [{"$literal": llm_id} for llm_id in llm_ids],
                        ],
                    },
                ]
            }
        rewritten_query = Query(**{"$expr": expr})

        res = self.server.cost_query(
            CostQueryReq(
                project_id=self._project_id(),
                query=rewritten_query,
                offset=offset,
                limit=limit,
            )
        )
        return res.results

    ################# Object Saving ##################
    # `_save_object` is the top level entry point for saving data to the weave server.
    # `_save_nested_objects` is a recursive method to dispatch saving of nested objects.
    #  it is called by `_save_object` above, as well as `create_call` and `finish_call`
    #  since we don't save the entire dictionary, but rather want to save any nested objects
    # `_save_object_basic` is the lowest level object saving logic which:
    #  - serializes the object to json
    #  - calls the server to save the object
    #  - creates an ObjectRef and attaches it to the object
    # `_save_op` and `_save_table` are the sister functions to `_save_object_basic`
    #  but for Ops and Tables respectively.

    @trace_sentry.global_trace_sentry.watch()
    def _save_object(self, val: Any, name: str, branch: str = "latest") -> ObjectRef:
        """Save an object to the weave server and return it's Ref. This is the top
        level entry point for saving any data to the weave server. Importantly, it
        will also save all children objects that are "Refable".

        Args:
            val: The object to save.
            name: The name to save the object under.
            branch: The branch to save the object under. Defaults to "latest".

        Returns:
            An ObjectRef to the saved object.
        """
        if isinstance(val, WeaveTable):
            # TODO: Probably should error here
            pass

        # If it's an Op, use the Op saving logic
        if is_op(val):
            # TODO: Probably should call _save_op directly here (or error)
            pass

        # Step 1: Recursively save all nested objects
        self._save_nested_objects(val, name=name)

        # Step 2: Save the object itself
        return self._save_object_basic(val, name, branch)

    @trace_sentry.global_trace_sentry.watch()
    def _save_nested_objects(self, obj: Any, name: Optional[str] = None) -> Any:
        """Recursively visits all values, ensuring that any "Refable" objects are
        saved and reffed.
        As of this writing, the only "Refable" objects are instances of:
        - weave.flow.obj.Object
        - weave.trace.op.Op
        - weave.trace.Table
        - weave.trace.vals.WeaveTable
        This procedure is a bit complicated, so it is worth making the details explicit:
        1. If the `obj` value already has a `ref`:
            - If the ref is to the current project, do nothing.
            - If the ref is to a different project, remove it. (to avoid cross-project references)
        2. If the `obj` value can be "reffed" (according to one of the above cases), invoke
            the appropriate "save" function for that type of object, and attach the ref result to `obj`.
            - `_save_object_basic` (for `weave.flow.obj.Object` instances)
            - `_save_op` (for `weave.trace.op.Op` instances)
            - `_save_table` (for `weave.trace.Table` and `weave.trace.vals.WeaveTable` instances)
        3. Otherwise, traverse all values within `obj` recursively, applying the above logic to each value.
        Important notes to developers: This method does not return anything - it _mutates_ the
        values that it traverses (specifically, it attaches `ref` values to them)

        Important: This method calls low level save methods directly - causing network events. Until
        these are backgrounded, they should not be invoked from inside a critical path.
        """
        # Base case: if the object is already refed
        #  - if the ref is to a different project, remove it
        #  - if the ref is to the current project, do nothing
        if (ref := get_ref(obj)) is not None:
            if ALLOW_MIXED_PROJECT_REFS:
                return
            # Check if existing ref is to current project, if not,
            # remove the ref and recreate it in the current project
            if ref.project == self.project:
                return
            remove_ref(obj)
        # Must defer import here to avoid circular import
        from weave.flow.obj import Object

        # Case 1: Object:
        # Here we recurse into each of the properties of the object
        # and save them, and then save the object itself.
        if isinstance(obj, Object):
            obj_rec = pydantic_object_record(obj)
            for v in obj_rec.__dict__.values():
                self._save_nested_objects(v)
            ref = self._save_object_basic(obj_rec, name or get_obj_name(obj_rec))
            # Personally, i think we should be passing `obj` into _save_object_basic,
            # and letting `to_json` handle converting the pydantic object into a jsonable object
            # but that might have unintended consequences. As a result, we break the
            # typical pattern and explicitly set the ref here.
            set_ref(obj, ref)

        # Case 2: Op:
        # Here we save the op itself.
        elif is_op(obj):
            # Ref is attached in here
            self._save_op(obj)

        # Case 3: Table
        elif isinstance(obj, Table):
            self._save_table(obj)

        # Case 4: WeaveTable
        elif isinstance(obj, WeaveTable):
            self._save_table(obj)

        # Special case: Custom recursive handling for WeaveObject with rows
        # TODO: Kinda hacky way to dispatching Dataset with rows: Table
        elif isinstance(obj, WeaveObject) and hasattr(obj, "rows"):
            self._save_nested_objects(obj.rows)

        # Recursive traversal of other pydantic objects
        elif isinstance(obj, (pydantic.BaseModel, pydantic.v1.BaseModel)):
            obj_rec = pydantic_object_record(obj)
            for v in obj_rec.__dict__.values():
                self._save_nested_objects(v)

        # Recursive traversal of other dataclasses
        elif dataclasses.is_dataclass(obj) and not isinstance(obj, Ref):
            obj_rec = dataclass_object_record(obj)
            for v in obj_rec.__dict__.values():
                self._save_nested_objects(v)

        # Recursive traversal of python structures
        elif isinstance_namedtuple(obj):
            for v in obj._asdict().values():
                self._save_nested_objects(v)
        elif isinstance(obj, (list, tuple)):
            for v in obj:
                self._save_nested_objects(v)
        elif isinstance(obj, dict):
            for v in obj.values():
                self._save_nested_objects(v)

    @trace_sentry.global_trace_sentry.watch()
    def _save_object_basic(
        self, val: Any, name: Optional[str] = None, branch: str = "latest"
    ) -> ObjectRef:
        """Directly saves an object to the weave server and attach
        the ref to the object. This is the lowest level object saving logic.
        """
        # The WeaveTable case is special because object saving happens inside
        # _save_object_nested and it has a special table_ref -- skip it here.
        if getattr(val, "_is_dirty", False) and not isinstance(val, WeaveTable):
            val.ref = None

        is_opdef = is_op(val)
        orig_val = val
        val = map_to_refs(val)
        if isinstance(val, ObjectRef):
            return val

        # `to_json` is mostly fast, except for CustomWeaveTypes
        # which incur network costs to serialize the payload
        json_val = to_json(val, self._project_id(), self.server)

        if name is None:
            if json_val.get("_type") == "CustomWeaveType":
                custom_name = json_val.get("weave_type", {}).get("type")
                name = custom_name

        if name is None:
            raise ValueError("Name must be provided for object saving")

        name = sanitize_object_name(name)

        response = self.server.obj_create(
            ObjCreateReq(
                obj=ObjSchemaForInsert(
                    project_id=self.entity + "/" + self.project,
                    object_id=name,
                    val=json_val,
                )
            )
        )
        ref: Ref
        if is_opdef:
            ref = OpRef(self.entity, self.project, name, response.digest)
        else:
            ref = ObjectRef(self.entity, self.project, name, response.digest)

        # Attach the ref to the object
        try:
            set_ref(orig_val, ref)
        except:
            # Don't worry if we can't set the ref.
            # This can happen for primitive types that don't have __dict__
            pass

        return ref

    @trace_sentry.global_trace_sentry.watch()
    def _save_op(self, op: Op, name: Optional[str] = None) -> ObjectRef:
        """
        Saves an Op to the weave server and returns the Ref. This is the sister
        function to _save_object_basic, but for Ops
        """
        if op.ref is not None:
            return op.ref

        if name is None:
            name = op.name

        return self._save_object_basic(op, name)

    @trace_sentry.global_trace_sentry.watch()
    def _save_table(self, table: Table) -> TableRef:
        """Saves a Table to the weave server and returns the TableRef.
        This is the sister function to _save_object_basic but for Tables.
        """
        rows = to_json(table.rows, self._project_id(), self.server)
        response = self.server.table_create(
            TableCreateReq(
                table=TableSchemaForInsert(project_id=self._project_id(), rows=rows)
            )
        )
        row_digests: Optional[list[str]] = None
        # This check is needed because in older versions of
        # the trace server, this will come back as an empty list.
        # In these cases, we want to set row_digests to None so that
        # the WeaveTable knows that it needs to fetch the rows from the server.
        if len(response.row_digests) == len(table.rows):
            row_digests = response.row_digests
        table_ref = TableRef(
            entity=self.entity,
            project=self.project,
            digest=response.digest,
            row_digests=row_digests,
        )
        table.ref = table_ref

        if isinstance(table, WeaveTable):
            table.table_ref = table_ref

        return table_ref

    ################ Internal Helpers ################

    def _ref_is_own(self, ref: Ref) -> bool:
        return isinstance(ref, Ref)

    def _project_id(self) -> str:
        return f"{self.entity}/{self.project}"

    @trace_sentry.global_trace_sentry.watch()
    def _op_calls(self, op: Op) -> CallsIter:
        op_ref = get_ref(op)
        if op_ref is None:
            raise ValueError(f"Can't get runs for unpublished op: {op}")
        return self.get_calls(CallsFilter(op_names=[op_ref.uri()]))

    @trace_sentry.global_trace_sentry.watch()
    def _objects(self, filter: Optional[ObjectVersionFilter] = None) -> list[ObjSchema]:
        if not filter:
            filter = ObjectVersionFilter()
        else:
            filter = filter.model_copy()
        filter = typing.cast(ObjectVersionFilter, filter)
        filter.is_op = False

        response = self.server.objs_query(
            ObjQueryReq(
                project_id=self._project_id(),
                filter=filter,
            )
        )
        return response.objs

    @trace_sentry.global_trace_sentry.watch()
    def _set_call_display_name(
        self, call: Call, display_name: Optional[str] = None
    ) -> None:
        # Removing call display name, use "" for db representation
        if display_name is None:
            display_name = ""
        self.server.call_update(
            CallUpdateReq(
                project_id=self._project_id(),
                call_id=call.id,
                display_name=display_name,
            )
        )

    def _remove_call_display_name(self, call: Call) -> None:
        self._set_call_display_name(call, None)

    def _ref_input_to(self, ref: ref_base.Ref) -> Sequence[Call]:
        raise NotImplementedError()

    def _ref_value_input_to(self, ref: ref_base.Ref) -> list[Call]:
        raise NotImplementedError()

    def _ref_output_of(self, ref: ObjectRef) -> typing.Optional[Call]:
        raise NotImplementedError()

    def _op_runs(self, op_def: Op) -> Sequence[Call]:
        raise NotImplementedError()

    def _ref_uri(self, name: str, version: str, path: str) -> str:
        return ObjectRef(self.entity, self.project, name, version).uri()

    def _flush(self) -> None:
        # Used to wait until all currently enqueued jobs are processed
        self.async_job_queue.flush()
        if self._server_is_flushable:
            # We don't want to do an instance check here because it could
            # be susceptible to shutdown race conditions. So we save a boolean
            # _server_is_flushable and only call this if we know the server is
            # flushable. The # type: ignore is safe because we check the type
            # first.
            self.server.call_processor.wait_until_all_processed()  # type: ignore


def send_start_call(
    project_id: str,
    call_id: str,
    op_str: str,
    trace_id: str,
    started_at: datetime.datetime,
    display_name: Optional[str],
    parent_id: Optional[str],
    inputs_with_refs: dict[str, Any],
    attributes: dict[str, Any],
    current_wb_run_id: Optional[str],
    server: TraceServerInterface,
) -> None:
    inputs_json = to_json(inputs_with_refs, project_id, server)
    start = StartedCallSchemaForInsert(
        project_id=project_id,
        id=call_id,
        op_name=op_str,
        display_name=display_name,
        trace_id=trace_id,
        started_at=started_at,
        parent_id=parent_id,
        inputs=inputs_json,
        attributes=attributes,
        wb_run_id=current_wb_run_id,
    )
    server.call_start(CallStartReq(start=start))


def send_end_call(
    project_id: str,
    call_id: str,
    ended_at: datetime.datetime,
    output: Any,
    summary: dict[str, Any],
    exception_str: Optional[str],
    server: TraceServerInterface,
) -> None:
    output_json = to_json(output, project_id, server)
    server.call_end(
        CallEndReq(
            end=EndedCallSchemaForInsert(
                project_id=project_id,
                id=call_id,
                ended_at=ended_at,
                output=output_json,
                summary=summary,
                exception=exception_str,
            )
        )
    )


def safe_current_wb_run_id() -> Optional[str]:
    try:
        import wandb

        wandb_run = wandb.run
        if wandb_run is None:
            return None
        return f"{wandb_run.entity}/{wandb_run.project}/{wandb_run.id}"
    except ImportError:
        return None


def check_wandb_run_matches(
    wandb_run_id: Optional[str], weave_entity: str, weave_project: str
) -> None:
    if wandb_run_id:
        # ex: "entity/project/run_id"
        wandb_entity, wandb_project, _ = wandb_run_id.split("/")
        if wandb_entity != weave_entity or wandb_project != weave_project:
            raise ValueError(
                f'Project Mismatch: weave and wandb must be initialized using the same project. Found wandb.init targeting project "{wandb_entity}/{wandb_project}" and weave.init targeting project "{weave_entity}/{weave_project}". To fix, please use the same project for both library initializations.'
            )


def _build_anonymous_op(name: str, config: Optional[Dict] = None) -> Op:
    if config is None:

        def op_fn(*args, **kwargs):  # type: ignore
            # Code-capture unavailable for this op
            pass

    else:

        def op_fn(*args, **kwargs):  # type: ignore
            # Code-capture unavailable for this op
            op_config = config

    op_fn.__name__ = name
    op = op_deco(op_fn)
    op.name = name
    return op


REDACT_KEYS = (
    "api_key",
    "Authorization",
)
REDACTED_VALUE = "REDACTED"


def redact_sensitive_keys(obj: typing.Any) -> typing.Any:
    # We should NEVER mutate reffed objects.
    #
    # 1. This code builds new objects that no longer have refs
    # 2. Even if we did an in-place edit, that would invalidate the ref (since
    # the ref is to the object's digest)
    if get_ref(obj):
        return obj

    if isinstance(obj, dict):
        dict_res = {}
        for k, v in obj.items():
            if k in REDACT_KEYS:
                dict_res[k] = REDACTED_VALUE
            else:
                dict_res[k] = redact_sensitive_keys(v)
        return dict_res

    elif isinstance(obj, list):
        list_res = []
        for v in obj:
            list_res.append(redact_sensitive_keys(v))
        return list_res

    elif isinstance(obj, tuple):
        tuple_res = []
        for v in obj:
            tuple_res.append(redact_sensitive_keys(v))
        return tuple(tuple_res)

    return obj


def sanitize_object_name(name: str) -> str:
    # Replaces any non-alphanumeric characters with a single dash and removes
    # any leading or trailing dashes. This is more restrictive than the DB
    # constraints and can be relaxed if needed.
    res = re.sub(r"([._-]{2,})+", "-", re.sub(r"[^\w._]+", "-", name)).strip("-_")
    if not res:
        raise ValueError(f"Invalid object name: {name}")
    if len(res) > 128:
        res = res[:128]
    return res


__docspec__ = [WeaveClient, Call, CallsIter]



import json
import typing
import copy

import contextvars
import contextlib

# This file contains a context manager for managing the state of the gql json cache.
# The GQL client automatically deserializes json responses into python objects.
# In weave, we type and store arbitrary json blobs as strings. This context manager
# allows us to use cache the deserialized python objects that the GQL client automatically
# produces, so that we can avoid deserializing the same json blob multiple times.


# This context var is used to store the json cache. It maps the serialized json string
# to the deserialized python object.
_GQL_JSON_CACHE: contextvars.ContextVar[dict[str, typing.Any]] = contextvars.ContextVar(
    "gql_json_cache", default={}
)


@contextlib.contextmanager
def gql_json_cache_context() -> typing.Iterator[None]:
    token = _GQL_JSON_CACHE.set({})
    try:
        yield
    finally:
        _GQL_JSON_CACHE.get().clear()
        _GQL_JSON_CACHE.reset(token)


def gql_json_cache() -> dict[str, typing.Any]:
    return _GQL_JSON_CACHE.get()


# This is the entry point for reading from the json cache.
# If the key is not in the cache, it will be deserialized and added to the cache.
def use_json(key: str) -> typing.Any:
    cache = gql_json_cache()
    result = cache.get(key, None)
    if result is None:
        result = json.loads(key)
        cache_json(key, result)
    return result


# This should only be called in two places. In use_json above,
# and in mappers_gql.py in the GQLStringToString mapper.
def cache_json(key: str, value: typing.Any) -> None:
    gql_json_cache()[key] = value



import typing

from weave.legacy.weave import graph, infer_types, weave_types


def test_node_with_generic():
    T = typing.TypeVar("T")
    t = graph.Node[T]
    assert infer_types.python_type_to_type(t) == weave_types.Function(
        {}, weave_types.Any()
    )


def test_with_literal():
    assert infer_types.python_type_to_type(
        typing.Literal["a", "b"]
    ) == weave_types.UnionType(
        weave_types.Const(weave_types.String(), "a"),
        weave_types.Const(weave_types.String(), "b"),
    )



import dataclasses
import json
import logging
import typing

import pyarrow as pa

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op
from weave.legacy.weave import (
    engine_trace,
    artifact_base,
    artifact_fs,
    artifact_mem,
    gql_json_cache,
    io_service,
)
from weave.legacy.weave.arrow import convert
from weave.legacy.weave.arrow.list_ import (
    ArrowWeaveList,
    ArrowWeaveListType,
    PathItemType,
    PathType,
    make_vec_none,
    weave_arrow_type_check,
)
from weave.legacy.weave.gql_op_plugin import wb_gql_op_plugin
from weave.legacy.weave.op_def import map_type
from weave.legacy.weave.ops_domain import trace_tree, wb_util, wbmedia
from weave.legacy.weave.ops_domain import wb_domain_types as wdt
from weave.legacy.weave.ops_domain.run_history import history_op_common
from weave.legacy.weave.ops_domain.run_history.context import (
    get_error_on_non_vectorized_history_transform,
)
from weave.legacy.weave.ops_domain.table import _patch_legacy_image_file_types
from weave.legacy.weave.wandb_interface import wandb_stream_table

tracer = engine_trace.tracer()


@op(
    render_info={"type": "function"},
    plugins=wb_gql_op_plugin(lambda inputs, inner: "historyKeys"),
    hidden=True,
)
def refine_history3_type(run: wdt.Run) -> types.Type:
    # TODO: Consider merging `_unflatten_history_object_type` into the main path
    return ArrowWeaveListType(
        _unflatten_history_object_type(history_op_common.refine_history_type(run))
    )


@op(
    render_info={"type": "function"},
    plugins=wb_gql_op_plugin(lambda inputs, inner: "historyKeys"),
    hidden=True,
)
def refine_history3_with_columns_type(
    run: wdt.Run, history_cols: list[str]
) -> types.Type:
    # TODO: Consider merging `_unflatten_history_object_type` into the main path
    return ArrowWeaveListType(
        _unflatten_history_object_type(
            history_op_common.refine_history_type(
                run,
                columns=history_op_common.get_full_columns_prefixed(run, history_cols),
            )
        )
    )


@op(
    name="run-history3_with_columns",
    refine_output_type=refine_history3_with_columns_type,
    plugins=wb_gql_op_plugin(history_op_common.make_run_history_gql_field),
    output_type=ArrowWeaveListType(types.TypedDict({})),
    hidden=True,
)
def history3_with_columns(run: wdt.Run, history_cols: list[str]):
    return _get_history3(
        run, history_op_common.get_full_columns_prefixed(run, history_cols)
    )


@op(
    name="run-history3",
    refine_output_type=refine_history3_type,
    plugins=wb_gql_op_plugin(history_op_common.make_run_history_gql_field),
    output_type=ArrowWeaveListType(types.TypedDict({})),
    hidden=True,
)
def history3(run: wdt.Run):
    # TODO: This is now equivalent to hist2
    return history_op_common.mock_history_rows(run)


@dataclasses.dataclass
class PathTree:
    children: typing.Dict[PathItemType, "PathTree"] = dataclasses.field(
        default_factory=dict
    )
    data: typing.Optional[typing.Any] = None


def _use_fast_path(flattened_object_type: types.Type) -> bool:
    if isinstance(flattened_object_type, types.TypedDict):
        for k, v in flattened_object_type.property_types.items():
            if not _use_fast_path(v):
                return False
        return True
    elif isinstance(flattened_object_type, types.List):
        return _use_fast_path(flattened_object_type.object_type)
    elif isinstance(flattened_object_type, types.UnionType):
        return all([_use_fast_path(m) for m in flattened_object_type.members])
    elif isinstance(
        flattened_object_type,
        (
            # If the table consists of only these leaf types, then we can use the fast path
            # If you find that you need to expand this list, please talk to
            # Shawn first.
            types.RefType,
            types.NoneType,
            types.Number,
            types.String,
            types.Boolean,
            types.Timestamp,
        ),
    ):
        return True

    # Encountered a type that requires the legacy path
    return False


def _fast_history3_concat(
    raw_history_awl_tables: list[pa.Table],
    live_data: list[dict],
) -> ArrowWeaveList:
    # This is the fast loading path. We only have the types contained in _use_fast_path
    # above.

    # This should be the only path used by StreamTable uses! We don't let users
    # put in the old wandb media types.

    # This path is pure arrow columnar operations, and should never do any
    # kind of python iteration.

    # Note we don't rely on the computed history type. We figure out the type
    # from the arrow types.

    all_awls: list[ArrowWeaveList] = []
    for table in raw_history_awl_tables:
        all_awls.append(ArrowWeaveList(table))

    # OK I lied, to_arrow does python iteration at the moment, but this is only
    # on the live set. If the live set remains smallish (which we can tune)
    # _fast_path should remain fast.
    all_awls.append(convert.to_arrow(live_data))

    # Nothing below here should do python iteration.

    all_converted_awls = []
    for awl in all_awls:

        def _convert_cols(
            awl: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if (
                isinstance(awl.object_type, types.TypedDict)
                and "_val" in awl.object_type.property_types
            ):
                # There are only two kinds of types that look like a _type, _val
                # struct: Timestamp and Ref. Happily, we can distinguish them by
                # looking at the type of the _val field
                val_field = awl._arrow_data.field("_val")
                if isinstance(val_field, (pa.DoubleArray, pa.Int64Array)):
                    # DoubleArray is the always included timestamp field
                    # Int64Array is a user defined timestamp.
                    return ArrowWeaveList(
                        val_field.cast("int64").cast(pa.timestamp("ms", tz="UTC")),
                        types.Timestamp(),
                        awl._artifact,
                    )
                elif isinstance(awl._arrow_data.field("_val"), pa.StringArray):
                    # StringArray is a Ref
                    return ArrowWeaveList(
                        val_field,
                        types.RefType(types.UnknownType()),
                        awl._artifact,
                    )
                else:
                    raise errors.WeaveWBHistoryTranslationError(
                        f"Encountered unexpected type for _val: {type(awl._arrow_data.field('_val'))}"
                    )
            return None

        awl = awl.map_column(fn=lambda x, y: x, pre_fn=_convert_cols)
        all_converted_awls.append(awl)

    return history_op_common.concat_awls(all_converted_awls)


def _check_fast_history3_concat_type(
    orig_expected_type: types.Type, fast_path_type: types.Type, depth: int = 0
) -> str:
    # Check if the output type of fast_history3_concat matches the expected type we get from
    # history type refinement.

    # This encodes business logic for places where we expect mismatches

    # Returns True if match

    if isinstance(orig_expected_type, types.UnionType) or isinstance(
        fast_path_type, types.UnionType
    ):
        # The new type may have nullable where the old didn't. Allow this, by just considering
        # both non_null types.
        orig_nullable, orig_non_null = types.split_none(orig_expected_type)
        fast_path_nullable, fast_non_null = types.split_none(fast_path_type)
        if orig_nullable or fast_path_nullable:
            return _check_fast_history3_concat_type(
                orig_non_null, fast_non_null, depth=depth
            )

    if isinstance(orig_expected_type, types.TypedDict) and isinstance(
        fast_path_type, types.TypedDict
    ):
        result = ""
        for k in set(orig_expected_type.property_types.keys()).union(
            fast_path_type.property_types.keys()
        ):
            sub_result = _check_fast_history3_concat_type(
                orig_expected_type.property_types[k],
                fast_path_type.property_types.get(k, types.NoneType()),
                depth=depth + 1,
            )
            if sub_result:
                result += (
                    "  " * depth + f"TypedDict property {k} mismatch:\n" + sub_result
                )
        return result
    elif isinstance(orig_expected_type, types.List) and isinstance(
        fast_path_type, types.List
    ):
        result = _check_fast_history3_concat_type(
            orig_expected_type.object_type, fast_path_type.object_type, depth=depth + 1
        )
        if result:
            return "  " * depth + "List object type mismatch:\n" + result
        return ""

    elif isinstance(orig_expected_type, types.UnionType) and isinstance(
        fast_path_type, types.UnionType
    ):
        result = ""

        # Unions can be out of order, so we check each orig member against
        # each fast_path member
        # We also may have fewer members in the fast_path type, because Refs are always
        # UnknownType, where in the orig type current they may be more specific and not
        # merged into a single RefType.
        for i in range(len(orig_expected_type.members)):
            orig_member = orig_expected_type.members[i]
            for j in range(len(fast_path_type.members)):
                fast_member = fast_path_type.members[j]
                sub_result = _check_fast_history3_concat_type(
                    orig_member, fast_member, depth=depth + 1
                )
                if not sub_result:
                    # A match, go to next orig member
                    break
            else:
                # no remaining fast path member matched the orig member
                result += (
                    "  " * depth
                    + f"Union type member mismatch, original_member: {orig_member} not found in remaining fast_path members\n"
                )
        return result
    elif isinstance(orig_expected_type, types.RefType) and isinstance(
        fast_path_type, types.RefType
    ):
        # The original type may have a full object_type, while in the fast path
        # we always return UnknownType as the object_type. This is expected.
        # So we just don't check object_type here.
        #
        # The original type will also be a more specific, like WandbArtifactRefType
        # but we only have RefType in the fast path. This is also expected.
        return ""
    else:
        if orig_expected_type.assign_type(fast_path_type):
            # For example, if we expect Number but got Float, Float is more specific
            # so that's OK.
            return ""
        if orig_expected_type != fast_path_type:
            orig_expected_str = str(orig_expected_type)
            if len(orig_expected_str) > 50:
                orig_expected_str = orig_expected_str[:50] + "..."
            fast_path_str = str(fast_path_type)
            if len(fast_path_str) > 50:
                fast_path_str = fast_path_str[:50] + "..."
            return (
                "  " * depth + f"Expected {orig_expected_str} but got {fast_path_str}\n"
            )
        return ""


def _get_history3(run: wdt.Run, columns=None):
    # 1. Get the flattened Weave-Type given HistoryKeys
    # 2. Read in the live set
    # 3. Raw-load each parquet file
    # 4. For each flattened column:
    #   a. If it contains a type requiring in-mem processing, then
    #       i. For liveset -> we must pass each item `_process_run_dict_item`, then use convert.to_arrow
    #       ii. For parquet -> we must extract the pyobj, us `_process_run_dict_item`, convert.to_arrow
    #   b. If the nested type contains a custom type or object type, then we must adjust the structure:
    #       i. For liveset -> we can pass each item to the simplified `_reconstruct_run_item`, then use convert.to_arrow(already_mapped = true)
    #       ii. For parquet -> we need to use the `_correct_awl` to adjust the structure.
    #   c. For all other types
    #       i. For liveset -> we can just use convert.to_arrow
    #       ii. For parquet -> no op
    # 5. Now we concat the converted liveset and parquet files
    # 6. Finally, unflatten the columns and return the result
    # 7. Optionally: verify the AWL

    artifact = artifact_mem.MemArtifact()

    # 1. Get the flattened Weave-Type given HistoryKeys
    flattened_object_type = history_op_common.refine_history_type(run, columns=columns)
    final_type = _unflatten_history_object_type(flattened_object_type)

    # 2. Read in the live set
    raw_live_data = _get_live_data_from_run(run, columns=columns)

    # 3.a: Raw-load each parquet file
    raw_history_awl_tables = _read_raw_history_awl_tables(
        run, columns=columns, artifact=artifact
    )

    # 3.b: Collapse unions
    union_collapsed_awl_tables = [
        _collapse_unions(awl) for awl in raw_history_awl_tables
    ]

    # 4. Process all the data
    raw_history_pa_tables = [
        history_op_common.awl_to_pa_table(awl) for awl in union_collapsed_awl_tables
    ]

    # 5 Now we concat the converted liveset and parquet files
    use_fast_path = _use_fast_path(flattened_object_type)
    if use_fast_path:
        concatted_awl = _fast_history3_concat(raw_history_pa_tables, raw_live_data)
        if len(concatted_awl) == 0:
            return convert.to_arrow([], types.List(final_type), artifact=artifact)
        if not isinstance(concatted_awl.object_type, types.TypedDict):
            raise errors.WeaveWBHistoryTranslationError(
                f"Expected fast_path object_type to be TypedDict, got {concatted_awl.object_type}"
            )

        # Need to get a new final_type, since we don't rely on the flattend_object_type
        # in the fast path.
        new_final_type = _unflatten_history_object_type(concatted_awl.object_type)
        type_mismatch_reason = _check_fast_history3_concat_type(
            final_type, new_final_type
        )
        if type_mismatch_reason:
            # Leaving this in to fire if we get an unexpected type from the fast
            # path. It's probably actually ok if this happens, and we could remove
            # these checks. But I want to understand scenarios in which it happens,
            # so raising here for now. - Shawn
            # raise errors.WeaveWBHistoryTranslationError(
            #     f"Fast path history3 concat produced unexpected type: {type_mismatch_reason}"
            # )
            pass
        final_type = new_final_type
    else:
        # Legacy slow path. This path drops down to python iteration in many
        # cases.
        logging.warning("Using legacy slow path for history3")
        run_path = wb_util.RunPath(
            run["project"]["entity"]["name"],
            run["project"]["name"],
            run["name"],
        )
        (
            live_columns,
            live_columns_already_mapped,
            processed_history_pa_tables,
        ) = _process_all_columns(
            flattened_object_type,
            raw_live_data,
            raw_history_pa_tables,
            run_path,
            artifact,
        )

        live_data_awl = _construct_live_data_awl(
            live_columns,
            live_columns_already_mapped,
            flattened_object_type,
            len(raw_live_data),
            artifact,
        )

        concatted_awl = history_op_common.concat_awls(
            [
                live_data_awl,
                *{
                    ArrowWeaveList(
                        table, object_type=flattened_object_type, artifact=artifact
                    )
                    for table in processed_history_pa_tables
                },
            ]
        )

        if len(concatted_awl) == 0:
            return convert.to_arrow([], types.List(final_type), artifact=artifact)

    sorted_table = history_op_common.sort_history_pa_table(
        history_op_common.awl_to_pa_table(concatted_awl)
    )

    # 6. Finally, unflatten the columns
    final_array = _unflatten_pa_table(sorted_table)

    # 7. Optionally: verify the AWL
    reason = weave_arrow_type_check(final_type, final_array)

    if reason != None:
        raise errors.WeaveWBHistoryTranslationError(
            f"Failed to effectively convert column of Gorilla Parquet History to expected history type: {reason}"
        )
    return ArrowWeaveList(
        final_array,
        final_type,
        artifact=artifact,
    )


def _construct_live_data_awl(
    live_columns: dict[str, list],
    live_columns_already_mapped: dict[str, list],
    flattened_object_type: types.TypedDict,
    num_rows: int,
    artifact: artifact_mem.MemArtifact,
) -> ArrowWeaveList:
    live_columns_to_data = [
        {k: v[i] for k, v in live_columns.items()} for i in range(num_rows)
    ]

    live_columns_already_mapped_to_data = [
        {k: v[i] for k, v in live_columns_already_mapped.items()}
        for i in range(num_rows)
    ]

    partial_live_data_awl = convert.to_arrow(
        live_columns_to_data,
        types.List(
            types.TypedDict(
                {
                    k: v
                    for k, v in flattened_object_type.property_types.items()
                    if k in live_columns
                }
            )
        ),
        artifact=artifact,
    )
    partial_live_data_already_mapped_awl = convert.to_arrow_from_list_and_artifact(
        live_columns_already_mapped_to_data,
        types.TypedDict(
            {
                k: v
                for k, v in flattened_object_type.property_types.items()
                if k in live_columns_already_mapped
            }
        ),
        artifact=artifact,
        py_objs_already_mapped=True,
    )

    field_names = []
    field_arrays = []
    for field_ndx, field in enumerate(partial_live_data_awl._arrow_data.type):
        field_names.append(field.name)
        field_arrays.append(partial_live_data_awl._arrow_data.field(field_ndx))

    for field_ndx, field in enumerate(
        partial_live_data_already_mapped_awl._arrow_data.type
    ):
        field_names.append(field.name)
        field_arrays.append(
            partial_live_data_already_mapped_awl._arrow_data.field(field_ndx)
        )

    return ArrowWeaveList(
        pa.StructArray.from_arrays(field_arrays, field_names),
        flattened_object_type,
        artifact=artifact,
    )


def _pa_table_has_column(pa_table: pa.Table, col_name: str) -> bool:
    return col_name in pa_table.column_names


def _create_empty_column(
    table: pa.Table, col_type: types.Type, artifact: artifact_base.Artifact
) -> pa.Array:
    return convert.to_arrow(
        [None] * len(table), types.List(col_type), artifact=artifact
    )._arrow_data


def _update_pa_table_column(
    table: pa.Table, col_name: str, arrow_col: pa.Array
) -> pa.Table:
    fields = [field.name for field in table.schema]
    return table.set_column(fields.index(col_name), col_name, arrow_col)


def _new_pa_table_column(
    table: pa.Table, col_name: str, arrow_col: pa.Array
) -> pa.Table:
    return table.append_column(col_name, arrow_col)


def _process_all_columns(
    flattened_object_type: types.TypedDict,
    raw_live_data: list[dict],
    raw_history_pa_tables: list[pa.Table],
    run_path: wb_util.RunPath,
    artifact: artifact_mem.MemArtifact,
):
    live_columns = {}
    live_columns_already_mapped = {}
    processed_history_pa_tables = [*raw_history_pa_tables]

    for col_name, col_type in flattened_object_type.property_types.items():
        raw_live_column = _extract_column_from_live_data(raw_live_data, col_name)
        processed_live_column = None
        if _column_contains_legacy_media_image(col_type):
            # If the column contains an image-file type, then we need to figure
            # out a more specific type for the column based on the first few
            # rows. This is an unfortunate side-effect of the fact that we don't
            # store image annotation metadata in the type system. See
            # `_patch_legacy_image_file_types` for more details on how we handle
            # this for older tables. This operation must be done on the liveset
            # and the history parquet tables, and the underlying type must be
            # updated accordingly. Not only that, but mask & box data are
            # variadic in type and sometimes require reading more files from
            # disk to understand the type. This is basically a dead end here -
            # unless we want to literally read the entire history parquet file
            # into memory, and then download all the image annotation data... We
            # have to then propagate back all the type changes... This is
            # horribly inefficient and we should probably just move on from
            # legacy image files
            pass
        if _column_type_requires_in_memory_transformation(col_type):
            _non_vectorized_warning(
                f"Encountered a history column requiring non-vectorized, in-memory processing: {col_name}: {col_type}"
            )
            processed_live_column = _process_column_in_memory(raw_live_column, run_path)
            live_columns[col_name] = processed_live_column

            for table_ndx, table in enumerate(processed_history_pa_tables):
                if not _pa_table_has_column(table, col_name):
                    new_table = _new_pa_table_column(
                        table, col_name, _create_empty_column(table, col_type, artifact)
                    )
                else:
                    new_table = _update_pa_table_column(
                        table,
                        col_name,
                        _process_history_column_in_memory(
                            table[col_name], col_type, run_path, artifact
                        ),
                    )
                processed_history_pa_tables[table_ndx] = new_table

        elif _is_directly_convertible_type(col_type):
            live_columns_already_mapped[col_name] = raw_live_column
            # Important that this runs before the else branch
            continue
        else:
            processed_live_column = _reconstruct_original_live_data_col(raw_live_column)
            live_columns_already_mapped[col_name] = processed_live_column

            for table_ndx, table in enumerate(processed_history_pa_tables):
                if not _pa_table_has_column(table, col_name):
                    new_table = _new_pa_table_column(
                        table, col_name, _create_empty_column(table, col_type, artifact)
                    )
                else:
                    new_col = _drop_types_from_encoded_types(
                        ArrowWeaveList(
                            table[col_name],
                            None,
                            artifact=artifact,
                        )
                    )
                    arrow_col = new_col._arrow_data
                    if types.Timestamp().assign_type(types.non_none(col_type)):
                        arrow_col = arrow_col.cast("int64").cast(
                            pa.timestamp("ms", tz="UTC")
                        )
                    new_table = _update_pa_table_column(table, col_name, arrow_col)
                processed_history_pa_tables[table_ndx] = new_table

    return live_columns, live_columns_already_mapped, processed_history_pa_tables


def _non_vectorized_warning(message: str):
    logging.warning(message)
    if get_error_on_non_vectorized_history_transform():
        raise errors.WeaveWBHistoryTranslationError(message)


def _build_array_from_tree(tree: PathTree) -> pa.Array:
    children_struct_array = []
    if len(tree.children) > 0:
        mask = None
        children_data = {}
        for k, v in tree.children.items():
            children_data[k] = _build_array_from_tree(v)
            child_is_null = _is_null(children_data[k])
            if mask is None:
                mask = child_is_null
            else:
                mask = pa.compute.and_(mask, child_is_null)

        children_struct = pa.StructArray.from_arrays(
            children_data.values(), children_data.keys(), mask=mask
        )
        children_struct_array = [children_struct]
    children_arrays = [*(tree.data or []), *children_struct_array]
    if len(children_arrays) == 0:
        raise errors.WeaveWBHistoryTranslationError(
            "Cannot build array from empty tree"
        )

    children_arrays_nonnull = [
        x for x in children_arrays if not isinstance(x, pa.NullArray)
    ]
    if not children_arrays_nonnull:
        children_arrays = [children_arrays[0]]
    else:
        children_arrays = children_arrays_nonnull

    if len(children_arrays) == 1:
        return children_arrays[0]

    num_rows = len(children_arrays[0])

    return _union_from_column_data(num_rows, children_arrays)


def _unflatten_pa_table(array: pa.Table) -> pa.StructArray:
    cols = array.schema.names
    new_tree = PathTree(data=[])
    for col_name in cols:
        path = col_name.split(".")
        target = new_tree
        for part in path:
            if part not in target.children:
                target.children[part] = PathTree(data=[])
            target = target.children[part]
        if not isinstance(target.data, list):
            raise errors.WeaveWBHistoryTranslationError(
                f"Encountered unexpected data in PathTree: {type(target.data)}"
            )
        target.data.append(array[col_name].combine_chunks())

    # Recursively resolve the tree
    return _build_array_from_tree(new_tree)


def _union_collapse_mapper(
    awl: ArrowWeaveList, path: PathType
) -> typing.Optional[ArrowWeaveList]:
    object_type = types.non_none(awl.object_type)
    if types.TypedDict({}).assign_type(object_type):
        object_type = typing.cast(types.TypedDict, object_type)
        columns = list(object_type.property_types.keys())
        arrow_data = awl._arrow_data
        arrow_columns = [c.name for c in arrow_data.type]
        if all([c.startswith("_type_") for c in columns]):
            logging.warning(
                f"Encountered history store union: this requires a non-vectorized transformation on the order of number of rows (but fortunately does not require transforming to python). Path: {path}, Type: {object_type}"
            )
            num_rows = len(arrow_data)
            weave_type_members = []
            for col_name in columns:
                weave_type_members.append(object_type.property_types[col_name])

            union_arr = _union_from_column_data(
                num_rows,
                [
                    arrow_data.field(arrow_columns.index(col_name))
                    for col_name in columns
                ],
            )
            return ArrowWeaveList(
                union_arr, types.union(*weave_type_members), awl._artifact
            )

    return None


def _union_from_column_data(num_rows: int, columns: list[pa.Array]) -> pa.Array:
    type_array = pa.nulls(num_rows, type=pa.int8())
    offset_array = pa.nulls(num_rows, type=pa.int32())
    arrays = []
    for col_ndx, column_data in enumerate(columns):
        if col_ndx == 0:
            # We always take the entire first column, then replace with mask
            # each subsequent column. This approach ensures that rows which have
            # nulls for every single column can be properly represented (ie. by the
            # first column)
            new_data = column_data
            arrays.append(new_data)
            type_array = pa.array([col_ndx] * len(new_data), type=pa.int8())
            offset_array = pa.array(list(range(len(new_data))), type=pa.int32())
            continue
        is_usable = pa.compute.invert(_is_null(column_data))
        new_data = column_data.filter(is_usable)
        arrays.append(new_data)
        type_array = pa.compute.replace_with_mask(
            type_array,
            is_usable,
            pa.array([col_ndx] * len(is_usable), type=pa.int8()),
        )
        offset_array = pa.compute.replace_with_mask(
            offset_array,
            is_usable,
            pa.array(list(range(len(new_data))), type=pa.int32()),
        )

    # TODO: This code tries to add a null we can use, but it makes the
    # union invalid against the weave type. Instead we can just use the first
    # array in the union.
    # TODO: Need to fix this code, it is not correct now, just hacked in the branch
    # arrays.append(pa.nulls(1, type=pa.int8()))
    #
    # Following lines are from weaveflow merge, but i believe this is already fixed
    #
    # type_array = pa.compute.fill_null(type_array, len(arrays) - 1)
    # offset_array = pa.compute.fill_null(offset_array, 0)

    return pa.UnionArray.from_dense(type_array, offset_array, arrays)


def _is_null(array: pa.Array) -> pa.Array:
    if isinstance(array, pa.ChunkedArray):
        array = array.combine_chunks()

    base_truth = pa.compute.is_null(array)

    if pa.types.is_struct(array.type):
        children_nulls = [_is_null(array.field(i)) for i in range(len(array.type))]
        if len(children_nulls) == 0:
            return pa.array([True] * len(array), type=pa.bool_())
        else:
            curr = children_nulls[0]
            for child in children_nulls[1:]:
                curr = pa.compute.and_(curr, child)
            return pa.compute.or_(curr, base_truth)

    return base_truth


def _collapse_unions(awl: ArrowWeaveList) -> ArrowWeaveList:
    return awl.map_column(_union_collapse_mapper)


def _drop_types_mapper(
    awl: ArrowWeaveList, path: PathType
) -> typing.Optional[ArrowWeaveList]:
    object_type = types.non_none(awl.object_type)
    if types.TypedDict(
        {"_type": types.optional(types.String()), "_val": types.Any()}
    ).assign_type(object_type):
        return awl.column("_val")
    return None


def _drop_types_from_encoded_types(awl: ArrowWeaveList) -> ArrowWeaveList:
    return awl.map_column(_drop_types_mapper)


def _get_live_data_from_run(run: wdt.Run, columns=None):
    raw_live_data = run["sampledParquetHistory"]["liveData"]
    if columns is None:
        return raw_live_data
    column_set = set(columns)
    return [
        {k: v for k, v in row.items() if k in column_set}
        for row in gql_json_cache.use_json(raw_live_data)
    ]


def _extract_column_from_live_data(live_data: list[dict], column_name: str):
    return [row.get(column_name, None) for row in live_data]


def _process_column_in_memory(live_column: list, run_path: wb_util.RunPath) -> list:
    return [_process_live_object_in_memory(item, run_path) for item in live_column]


def _process_live_object_in_memory(
    live_data: typing.Any, run_path: wb_util.RunPath
) -> typing.Any:
    if isinstance(live_data, list):
        return [_process_live_object_in_memory(cell, run_path) for cell in live_data]
    if isinstance(live_data, dict):
        if live_data.get("_type") != None:
            return wb_util._process_run_dict_item(live_data, run_path)
        else:
            return {
                key: _process_live_object_in_memory(val, run_path)
                for key, val in live_data.items()
            }
    return live_data


def _process_history_column_in_memory(
    history_column: pa.Array,
    col_type: types.Type,
    run_path: wb_util.RunPath,
    artifact: artifact_base.Artifact,
) -> pa.Array:
    pyobj = history_column.to_pylist()
    processed_data = _process_column_in_memory(pyobj, run_path)
    awl = convert.to_arrow(
        processed_data,
        types.List(col_type),
        artifact=artifact,
    )
    return pa.chunked_array([awl._arrow_data])


# Copy from common - need to merge back
def _read_raw_history_awl_tables(
    run: wdt.Run,
    columns=None,
    artifact: typing.Optional[artifact_base.Artifact] = None,
) -> list[ArrowWeaveList]:
    io = io_service.get_sync_client()
    tables = []
    for url in run["sampledParquetHistory"]["parquetUrls"]:
        local_path = io.ensure_file_downloaded(url)
        if local_path is not None:
            path = io.fs.path(local_path)
            awl = history_op_common.awl_from_local_parquet_path(
                path, None, columns=columns, artifact=artifact
            )
            awl = awl.map_column(_parse_bytes_mapper)
            tables.append(awl)
    return tables

    # return history_op_common.process_history_awl_tables(tables)


def _parse_bytes_to_json(bytes: bytes):
    return json.loads(bytes.decode("utf-8"))


def _parse_bytes_mapper(
    awl: ArrowWeaveList, path: PathType
) -> typing.Optional[ArrowWeaveList]:
    obj_type = types.non_none(awl.object_type)
    if types.Bytes().assign_type(obj_type):
        if len(awl._arrow_data) == awl._arrow_data.null_count:
            return make_vec_none(len(awl._arrow_data))
        _non_vectorized_warning(
            f"Encountered bytes in column {path}, requires in-memory processing"
        )
        data = [
            _parse_bytes_to_json(row) if row is not None else None
            for row in awl._arrow_data.to_pylist()
        ]
        wb_type = types.TypeRegistry.type_of(data)
        if not hasattr(wb_type, "object_type"):
            raise errors.WeaveWBHistoryTranslationError(
                f"Expected type with object_type attribute, got {wb_type}"
            )
        wb_type = typing.cast(types.List, wb_type)
        return convert.to_arrow_from_list_and_artifact(
            data,
            wb_type.object_type,
            artifact=awl._artifact or artifact_mem.MemArtifact(),
            py_objs_already_mapped=True,
        )
    return None


def _column_type_requires_in_memory_transformation(col_type: types.Type):
    return _column_type_requires_in_memory_transformation_recursive(col_type)


def _column_contains_legacy_media_image(col_type: types.Type):
    return _column_contains_legacy_media_image_recursive(col_type)


def _is_directly_convertible_type(col_type: types.Type):
    return _is_directly_convertible_type_recursive(col_type)


_weave_types_requiring_in_memory_transformation = (
    # TODO: We should be able to move some (or all?) of these to
    # a vectorized approach. At a minimum we should be able to do
    # this for ImageArtifactFileRefType - similar to how we do it
    # in history2.
    wbmedia.ImageArtifactFileRefType,  # type: ignore
    wbmedia.AudioArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.BokehArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.VideoArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.Object3DArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.MoleculeArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.HtmlArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.LegacyTableNDArrayType,
    wb_util.WbHistogram,
    trace_tree.WBTraceTree.WeaveType,  # type: ignore
    types.UnknownType,
    artifact_fs.FilesystemArtifactFileType,  # Tables
)


def _column_type_requires_in_memory_transformation_recursive(col_type: types.Type):
    if types.NoneType().assign_type(col_type):
        return False
    non_none_type = types.non_none(col_type)
    if isinstance(non_none_type, types.UnionType):
        return True
    elif isinstance(non_none_type, _weave_types_requiring_in_memory_transformation):
        return True
    elif isinstance(non_none_type, types.List):
        return _column_type_requires_in_memory_transformation_recursive(
            non_none_type.object_type
        )
    elif isinstance(non_none_type, types.TypedDict):
        for k, v in non_none_type.property_types.items():
            if _column_type_requires_in_memory_transformation_recursive(v):
                return True
        return False
    else:
        return False


def _column_contains_legacy_media_image_recursive(col_type: types.Type):
    if types.NoneType().assign_type(col_type):
        return False
    non_none_type = types.non_none(col_type)
    if isinstance(non_none_type, types.UnionType):
        return True
    elif isinstance(non_none_type, wbmedia.ImageArtifactFileRefType):
        return True
    elif isinstance(non_none_type, types.List):
        return _column_type_requires_in_memory_transformation_recursive(
            non_none_type.object_type
        )
    elif isinstance(non_none_type, types.TypedDict):
        for k, v in non_none_type.property_types.items():
            if _column_type_requires_in_memory_transformation_recursive(v):
                return True
        return False
    else:
        return False


def _is_directly_convertible_type_recursive(col_type: types.Type):
    if types.NoneType().assign_type(col_type):
        return True
    non_none_type = types.non_none(col_type)
    if types.union(
        *[
            types.Number(),
            types.Int(),
            types.Float(),
            types.String(),
            types.Boolean(),
        ]
    ).assign_type(col_type):
        return True
    elif isinstance(non_none_type, types.List):
        return _is_directly_convertible_type_recursive(non_none_type.object_type)
    elif isinstance(non_none_type, types.TypedDict):
        for k, v in non_none_type.property_types.items():
            if not _is_directly_convertible_type_recursive(v):
                return False
        return True
    else:
        return False


def _reconstruct_original_live_data_col(row: list):
    return [_reconstruct_original_live_data_cell(cell) for cell in row]


def _reconstruct_original_live_data_cell(live_data: typing.Any) -> typing.Any:
    if isinstance(live_data, list):
        return [_reconstruct_original_live_data_cell(cell) for cell in live_data]
    if isinstance(live_data, dict):
        if wandb_stream_table.is_weave_encoded_history_cell(live_data):
            return live_data["_val"]
        return {
            key: _reconstruct_original_live_data_cell(val)
            for key, val in live_data.items()
        }
    return live_data


def _unflatten_history_object_type(obj_type: types.TypedDict) -> types.TypedDict:
    # Need to combine top-level maps. Note: this is not a recursive function
    # since the flattening only happens top level in gorilla.
    dict_keys: dict[str, types.Type] = {}
    for key, val in obj_type.property_types.items():
        path_parts = key.split(".")
        prop_types = dict_keys

        for part in path_parts[:-1]:
            found = False
            prop_types_inner: dict[str, types.Type] = {}
            default_dict = types.TypedDict(prop_types_inner)
            if part not in prop_types:
                prop_types[part] = default_dict
                prop_types = prop_types_inner
            elif part in prop_types:
                prop_type = prop_types[part]
                if isinstance(prop_type, types.TypedDict):
                    prop_types = prop_type.property_types
                    found = True
                elif isinstance(prop_type, types.UnionType):
                    for member in prop_type.members:
                        if isinstance(member, types.TypedDict):
                            prop_types = member.property_types
                            found = True
                            break

                if not found:
                    prop_types[part] = types.union(prop_type, default_dict)
                    prop_types = prop_types_inner

        final_part = path_parts[-1]
        if final_part not in prop_types:
            prop_types[final_part] = val
        else:
            prop_types[final_part] = types.union(val, prop_types[final_part])

    return types.TypedDict(dict_keys)



import weave


def test_simple_node():
    num = weave.save(5)
    # store a Node in a sub-key
    contains_node = weave.save({"node": num + 3})
    assert weave.use(contains_node["node"] + 9) == 17


def test_callable_output_type():
    typed_dict = weave.save({"a": {"b": 5}})
    contains_node = weave.save({"node": typed_dict["a"]})
    # calling ["b"] should succeed
    called = contains_node["node"]["b"]
    assert called.type == weave.types.Int()
    assert weave.use(called) == 5


def test_callable_input_type():
    list_ = weave.save([{"a": 5}, {"a": 7}])
    contains_node = weave.save({"node": list_.map(lambda r: r["a"])})
    # calling .map should succeed
    called = contains_node["node"].map(lambda r: r + 3)
    assert called.type == weave.types.List(weave.types.Number())
    assert weave.use(called) == [8, 10]



from weave.legacy.weave import context_state as _context

_loading_builtins_token = _context.set_loading_built_ins()

from weave.legacy.weave.ecosystem.wandb import wandb_objs
from weave.legacy.weave.ecosystem.wandb.panel_distribution import *
from weave.legacy.weave.ecosystem.wandb.panel_geo import *
from weave.legacy.weave.ecosystem.wandb.panel_scatter import *
from weave.legacy.weave.ecosystem.wandb.panel_time_series import *
from weave.legacy.weave.ecosystem.wandb.run_chain import *
from weave.legacy.weave.ecosystem.wandb.weave_plotly import *

_context.clear_loading_built_ins(_loading_builtins_token)



import dataclasses

import pytest

import weave
import weave.legacy.weave
import weave.legacy.weave.weave_types
from weave.legacy.weave import _dict_utils, runs
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType
from weave.legacy.weave.ops_domain import wbmedia

from ... import errors


def test_typeof_string():
    t = types.TypeRegistry.type_of("x")
    assert t == types.String()


def test_typeof_list_const_string():
    t = types.TypeRegistry.type_of(["x"])
    assert t == types.List(types.String())


def test_serialize_const_string():
    t = types.Const(types.String(), "x")
    ser = t.to_dict()
    deser = types.TypeRegistry.type_from_dict(ser)
    assert t == deser


def test_merge_typedict_keys_are_stable():
    for i in range(10):
        t = types.TypedDict(
            {"a": types.String(), "b": types.String(), "c": types.String()}
        )
        t2 = types.TypedDict(
            {"a": types.String(), "b": types.String(), "c": types.String()}
        )
        r = types.merge_types(t, t2)
        assert list(r.property_types.keys()) == ["a", "b", "c"]


def test_merge_through_tags():
    t = TaggedValueType(
        types.TypedDict({"tag": types.Number()}),
        types.TypedDict(
            {"a": types.String(), "b": types.Number(), "c": types.String()}
        ),
    )
    t2 = TaggedValueType(
        types.TypedDict({"tag": types.Number()}),
        types.TypedDict(
            {"a": types.String(), "b": types.String(), "d": types.String()}
        ),
    )
    r = types.merge_types(t, t2)
    correct_type = TaggedValueType(
        types.TypedDict(property_types={"tag": types.Number()}),
        types.TypedDict(
            property_types={
                "a": types.String(),
                "b": types.UnionType(types.Number(), types.String()),
                "c": types.UnionType(types.String(), types.NoneType()),
                "d": types.UnionType(types.NoneType(), types.String()),
            }
        ),
    )
    assert correct_type.assign_type(r)
    assert r.assign_type(correct_type)


def test_merge_tag_union_unknown():
    t = TaggedValueType(
        types.TypedDict(property_types={"run": types.Int()}),
        types.TypedDict(
            property_types={
                "step": types.UnionType(types.NoneType(), types.Float()),
                "prompt": types.UnionType(types.NoneType(), types.String()),
                "image": types.UnionType(types.String(), types.NoneType()),
            }
        ),
    )
    t2 = TaggedValueType(
        types.TypedDict(property_types={"run": types.Int()}),
        types.TypedDict(
            property_types={
                "step": types.UnknownType(),
                "prompt": types.UnknownType(),
                "image": types.UnknownType(),
            }
        ),
    )
    r = types.merge_types(t, t2)
    assert r == t


def test_tagged_unions_simple():
    assert weave.types.optional(weave.types.Int()).assign_type(
        TaggedValueType(
            types.TypedDict({"a": weave.types.String()}),
            weave.types.optional(weave.types.Int()),
        )
    )


def test_tag_assignment_through_union():
    base = TaggedValueType(types.TypedDict({"a": types.Number()}), types.Any())
    t = TaggedValueType(types.TypedDict({"a": types.Number()}), types.String())
    t2 = TaggedValueType(types.TypedDict({"a": types.Number()}), types.Number())
    union = types.union(t, t2)
    t3 = TaggedValueType(types.TypedDict({"b": types.Number()}), union)
    assert base.assign_type(t3)


def test_typeof_bool():
    assert types.TypeRegistry.type_of(False) == types.Boolean()


def test_typeof_type():
    assert types.TypeRegistry.type_of(types.Int()) == types.TypeType()


def test_type_tofromdict():
    d = types.TypeType().to_dict()
    assert types.TypeRegistry.type_from_dict(d) == types.TypeType()


def test_typeof_list_runs():
    l = [
        runs.Run("a", "op", inputs={"a": "x"}, output=4.9),
        runs.Run("b", "op", inputs={"a": "x", "b": 9}, output=3.3),
    ]
    actual = types.TypeRegistry.type_of(l)
    print("test_typeof_list_runs.actual", actual)

    assert actual == types.List(
        types.RunType(
            inputs=types.TypedDict(
                {"a": types.String(), "b": types.optional(types.Int())}
            ),
            history=types.List(types.UnknownType()),
            output=types.Float(),
        ),
    )


def test_typeof_list_dict_merge():
    d = [{"a": 6, "b": "x"}, {"a": 5, "b": None}]
    assert types.TypeRegistry.type_of(d) == types.List(
        types.TypedDict({"a": types.Int(), "b": types.optional(types.String())})
    )


def test_typeof_nested_dict_merge():
    """Tests that nested merging is disabled."""
    t1 = weave.legacy.weave.weave_types.TypedDict(
        {"a": weave.legacy.weave.weave_types.TypedDict({"b": types.Int()})}
    )
    t2 = weave.legacy.weave.weave_types.TypedDict(
        {"a": weave.legacy.weave.weave_types.TypedDict({"c": types.String()})}
    )
    merged_type = _dict_utils.typeddict_merge_output_type({"self": t1, "other": t2})
    assert merged_type == weave.legacy.weave.weave_types.TypedDict(
        {"a": weave.legacy.weave.weave_types.TypedDict({"c": types.String()})}
    )


def test_dict_without_key_type():
    fully_typed = weave.legacy.weave.weave_types.TypeRegistry.type_from_dict(
        {"type": "dict", "keyType": "string", "objectType": "number"}
    )
    partial_typed = weave.legacy.weave.weave_types.TypeRegistry.type_from_dict(
        {"type": "dict", "objectType": "number"}
    )
    assert fully_typed.assign_type(partial_typed)


# def test_union_unknown():
#     assert (
#         weave.legacy.weave.weave_types.union(
#             weave.legacy.weave.weave_types.String(), weave.legacy.weave.weave_types.UnknownType()
#         )
#         == weave.legacy.weave.weave_types.String()
#     )
#     assert (
#         weave.legacy.weave.weave_types.union(
#             weave.legacy.weave.weave_types.UnknownType(), weave.legacy.weave.weave_types.UnknownType()
#         )
#         == weave.legacy.weave.weave_types.UnknownType()
#     )
#     assert (
#         weave.legacy.weave.weave_types.union(
#             weave.legacy.weave.weave_types.UnknownType(),
#             weave.legacy.weave.weave_types.UnknownType(),
#             weave.types.String(),
#         )
#         == weave.legacy.weave.weave_types.String()
#     )


def test_union_access():
    ### Type return

    # Not all members have props
    unioned = weave.legacy.weave.weave_types.union(
        weave.legacy.weave.weave_types.String(),
        weave.legacy.weave.weave_types.List(weave.legacy.weave.weave_types.String()),
    )
    with pytest.raises(AttributeError):
        unioned.object_type

    # Combined dicts
    unioned = weave.legacy.weave.weave_types.union(
        weave.legacy.weave.weave_types.List(weave.legacy.weave.weave_types.String()),
        weave.legacy.weave.weave_types.List(weave.legacy.weave.weave_types.Number()),
    )
    assert unioned.object_type == weave.legacy.weave.weave_types.union(
        weave.legacy.weave.weave_types.String(), weave.legacy.weave.weave_types.Number()
    )

    # Nullable type
    unioned = weave.legacy.weave.weave_types.union(
        weave.legacy.weave.weave_types.NoneType(),
        weave.legacy.weave.weave_types.List(weave.legacy.weave.weave_types.String()),
    )
    assert unioned.object_type == weave.legacy.weave.weave_types.union(
        weave.legacy.weave.weave_types.String(), weave.legacy.weave.weave_types.NoneType()
    )

    ### Dict Return
    # Not all members have props
    unioned = weave.legacy.weave.weave_types.union(
        weave.legacy.weave.weave_types.String(),
        weave.legacy.weave.weave_types.TypedDict({"a": weave.legacy.weave.weave_types.String()}),
    )
    with pytest.raises(AttributeError):
        unioned.property_types

    # Combined dicts
    unioned = weave.legacy.weave.weave_types.union(
        weave.legacy.weave.weave_types.TypedDict(
            {
                "same": weave.legacy.weave.weave_types.Number(),
                "solo_a": weave.legacy.weave.weave_types.Number(),
                "differ": weave.legacy.weave.weave_types.Number(),
            }
        ),
        weave.legacy.weave.weave_types.TypedDict(
            {
                "same": weave.legacy.weave.weave_types.Number(),
                "solo_b": weave.legacy.weave.weave_types.String(),
                "differ": weave.legacy.weave.weave_types.String(),
            }
        ),
    )
    assert unioned.property_types == {
        "same": weave.legacy.weave.weave_types.Number(),
        "solo_a": weave.legacy.weave.weave_types.union(
            weave.legacy.weave.weave_types.Number(), weave.legacy.weave.weave_types.NoneType()
        ),
        "solo_b": weave.legacy.weave.weave_types.union(
            weave.legacy.weave.weave_types.String(), weave.legacy.weave.weave_types.NoneType()
        ),
        "differ": weave.legacy.weave.weave_types.union(
            weave.legacy.weave.weave_types.Number(), weave.legacy.weave.weave_types.String()
        ),
    }

    # Nullable type
    unioned = weave.legacy.weave.weave_types.union(
        weave.legacy.weave.weave_types.NoneType(),
        weave.legacy.weave.weave_types.TypedDict({"a": weave.legacy.weave.weave_types.String()}),
    )
    assert unioned.property_types == {
        "a": weave.legacy.weave.weave_types.union(
            weave.legacy.weave.weave_types.String(), weave.legacy.weave.weave_types.NoneType()
        )
    }


def test_typeof_node():
    n = weave.save(5)
    assert weave.type_of(n + 5) == types.Function({}, types.Number())


@dataclasses.dataclass(frozen=True)
class SublistType(types.Type):
    _base_type = types.List
    object_type: types.Type = types.UnknownType()


def test_subtype_list():
    assert types.List(types.Int()).assign_type(types.List(types.Int()))
    assert not types.List(types.String()).assign_type(types.List(types.Int()))

    assert types.List(types.Int()).assign_type(SublistType(types.Int()))
    assert not types.List(types.String()).assign_type(SublistType(types.Int()))


def test_typeddict_to_dict():
    assert types.Dict(types.String(), types.Int()).assign_type(
        types.TypedDict({"a": types.Int(), "b": types.Int()})
    )
    assert not types.Dict(types.String(), types.Int()).assign_type(
        types.TypedDict({"a": types.Int(), "b": types.String()})
    )


# The following test tests all permutionations of container types Consts, Tags,
# and Unions. Note: we might deicde to disallow consts of unions which would
# make some of these tests not applicable. However, this is exhastive given our
# current implementation.
@pytest.mark.parametrize(
    "in_type, out_type",
    [
        # Units
        (types.Number(), types.Number()),
        (types.NoneType(), types.Invalid()),
        # Const Units
        (
            types.Const(types.Number(), 5),
            types.Const(types.Number(), 5),
        ),
        (types.Const(types.NoneType(), None), types.Invalid()),
        # Tagged Units
        (
            TaggedValueType(types.TypedDict({}), types.Number()),
            TaggedValueType(types.TypedDict({}), types.Number()),
        ),
        (
            TaggedValueType(types.TypedDict({}), types.NoneType()),
            TaggedValueType(types.TypedDict({}), types.Invalid()),
        ),
        # Tagged Const Units
        (
            TaggedValueType(types.TypedDict({}), types.Const(types.Number(), 5)),
            TaggedValueType(types.TypedDict({}), types.Const(types.Number(), 5)),
        ),
        (
            TaggedValueType(types.TypedDict({}), types.Const(types.NoneType(), None)),
            TaggedValueType(types.TypedDict({}), types.Invalid()),
        ),
        # Const Tagged Units
        (
            types.Const(TaggedValueType(types.TypedDict({}), types.Number()), 5),
            types.Const(TaggedValueType(types.TypedDict({}), types.Number()), 5),
        ),
        (
            types.Const(TaggedValueType(types.TypedDict({}), types.NoneType()), None),
            TaggedValueType(types.TypedDict({}), types.Invalid()),
        ),
        # Union Units
        (types.union(types.NoneType(), types.Number()), types.Number()),
        (
            types.union(types.String(), types.Number()),
            types.union(types.String(), types.Number()),
        ),
        # Union Const Units
        (
            types.union(
                types.Const(types.NoneType(), None), types.Const(types.Number(), 5)
            ),
            types.Const(types.Number(), 5),
        ),
        (
            types.union(
                types.Const(types.String(), "hello"), types.Const(types.Number(), 5)
            ),
            types.union(
                types.Const(types.String(), "hello"), types.Const(types.Number(), 5)
            ),
        ),
        # Union Tagged Units
        (
            types.union(
                TaggedValueType(types.TypedDict({}), types.NoneType()),
                TaggedValueType(types.TypedDict({}), types.Number()),
            ),
            TaggedValueType(types.TypedDict({}), types.Number()),
        ),
        (
            types.union(
                TaggedValueType(types.TypedDict({}), types.String()),
                TaggedValueType(types.TypedDict({}), types.Number()),
            ),
            types.union(
                TaggedValueType(types.TypedDict({}), types.String()),
                TaggedValueType(types.TypedDict({}), types.Number()),
            ),
        ),
        # Union Tagged Consts Units
        (
            types.union(
                TaggedValueType(types.TypedDict({}), types.NoneType()),
                TaggedValueType(types.TypedDict({}), types.Const(types.Number(), 5)),
            ),
            TaggedValueType(types.TypedDict({}), types.Const(types.Number(), 5)),
        ),
        (
            types.union(
                TaggedValueType(
                    types.TypedDict({}), types.Const(types.String(), "hello")
                ),
                TaggedValueType(types.TypedDict({}), types.Const(types.Number(), 5)),
            ),
            types.union(
                TaggedValueType(
                    types.TypedDict({}), types.Const(types.String(), "hello")
                ),
                TaggedValueType(types.TypedDict({}), types.Const(types.Number(), 5)),
            ),
        ),
        # Tagged Union Units
        (
            TaggedValueType(
                types.TypedDict({}), types.union(types.NoneType(), types.Number())
            ),
            TaggedValueType(types.TypedDict({}), types.Number()),
        ),
        (
            TaggedValueType(
                types.TypedDict({}), types.union(types.String(), types.Number())
            ),
            TaggedValueType(
                types.TypedDict({}), types.union(types.String(), types.Number())
            ),
        ),
        # Union Tagged Union
        (
            types.UnionType(
                types.NoneType(),
                TaggedValueType(
                    types.TypedDict({}), types.union(types.NoneType(), types.Number())
                ),
            ),
            TaggedValueType(types.TypedDict({}), types.Number()),
        ),
        # Tagged Union Const Units
        (
            TaggedValueType(
                types.TypedDict({}),
                types.union(
                    types.Const(types.NoneType(), None), types.Const(types.Number(), 5)
                ),
            ),
            TaggedValueType(types.TypedDict({}), types.Const(types.Number(), 5)),
        ),
        (
            TaggedValueType(
                types.TypedDict({}),
                types.union(
                    types.Const(types.String(), "hello"), types.Const(types.Number(), 5)
                ),
            ),
            TaggedValueType(
                types.TypedDict({}),
                types.union(
                    types.Const(types.String(), "hello"), types.Const(types.Number(), 5)
                ),
            ),
        ),
        # Union with multiple non-none members
        (
            types.union(types.NoneType(), types.Number(), types.String()),
            types.union(types.Number(), types.String()),
        ),
        # Union with multiple none-like members
        (
            types.union(
                types.NoneType(), TaggedValueType(types.TypedDict({}), types.NoneType())
            ),
            types.Invalid(),
        ),
    ],
)
def test_non_none(in_type, out_type):
    assert types.non_none(in_type) == out_type


def test_const_union_resolves_union():
    assert (
        types.Const(types.union(types.NoneType(), types.Number()), 5).val_type
        == types.Int()
    )


def test_floatint_merged():
    assert weave.type_of([1.0, 2.0]).object_type == types.Float()
    assert weave.type_of([1.0, 2]).object_type == types.Float()
    assert weave.type_of([1, 2]).object_type == types.Int()


def test_typetype():
    tt = weave.type_of(weave.types.TypedDict({"a": weave.types.Int()}))
    assert tt == weave.types.TypeType(
        attr_types={
            "not_required_keys": weave.types.List(types.String()),
            "property_types": weave.types.Dict(types.String(), types.TypeType()),
        }
    )


def test_typetype_disjoint_from_normal_type_hierarchy():
    assert not weave.type_of(weave.types.List()).assign_type(weave.type_of(None))


def test_typetype_nodes():
    t = weave.save(weave.types.TypedDict({"a": weave.types.Int()}))
    assert weave.use(t.property_types) == {"a": weave.types.Int()}


def test_typetype_tofrom_dict():
    t = weave.type_of(types.TypedDict())
    d = t.to_dict()
    t2 = types.TypeType.from_dict(d)
    assert t == t2


def test_union_of_typetype_can_be_compared_to_other():
    t = types.UnionType(types.NoneType(), types.TypeType())
    assert t == t


def test_assign_dict_to_typeddict():
    assert weave.types.TypedDict({}).assign_type(
        weave.types.Dict(weave.types.String(), weave.types.String())
    )


def test_type_of_empty_array_union():
    assert weave.type_of(
        [
            {"a": []},
            {"a": [1]},
        ]
    ) == weave.types.List(
        weave.types.TypedDict({"a": weave.types.List(weave.types.Int())})
    )


def test_type_hash():
    assert hash(types.NoneType()) == hash(types.NoneType())
    assert hash(types.List(types.Int())) == hash(types.List(types.Int()))
    assert hash(types.TypedDict({"a": types.Int()})) == hash(
        types.TypedDict({"a": types.Int()})
    )
    assert hash(types.UnionType(types.NoneType(), types.String())) == hash(
        types.UnionType(types.String(), types.NoneType())
    )


def test_tagged_value_flow():
    vt_1 = types.TypedDict({"a": types.Int()})
    tt_1 = types.TypedDict({"b": types.String()})
    t_1 = TaggedValueType(vt_1, tt_1)

    vt_2 = types.TypedDict({"c": types.Boolean()})
    tt_2 = types.TypedDict({"d": types.Float()})
    t_2 = TaggedValueType(vt_2, tt_2)

    ut = types.TypedDict({"e": types.Timestamp()})
    uv = types.union(t_1, t_2)

    tut = TaggedValueType(ut, uv)

    assert tut.members == [TaggedValueType(ut, t_1), TaggedValueType(ut, t_2)]
    assert tut.property_types == {
        "d": TaggedValueType(ut, types.optional(TaggedValueType(vt_2, types.Float()))),
        "b": TaggedValueType(ut, types.optional(TaggedValueType(vt_1, types.String()))),
    }

    row_type = types.TypedDict({"a": types.Int(), "b": types.String()})
    list_type = types.List(row_type)
    list_tag_type = types.TypedDict({"t_1": types.Boolean()})
    tagged_list_type = TaggedValueType(list_tag_type, list_type)

    row_type_2 = types.TypedDict({"c": types.Int(), "b": types.String()})
    list_type_2 = types.List(row_type_2)
    list_tag_type_2 = types.TypedDict({"t_2": types.Boolean()})
    tagged_list_type_2 = TaggedValueType(list_tag_type_2, list_type_2)

    union_type = types.union(tagged_list_type, tagged_list_type_2)
    list_of_unions = types.List(union_type)

    pts = list_of_unions.object_type.object_type.property_types

    assert pts["a"] == types.optional(TaggedValueType(list_tag_type, types.Int()))
    assert pts["b"] == types.union(
        TaggedValueType(list_tag_type, types.String()),
        TaggedValueType(list_tag_type_2, types.String()),
    )
    assert pts["c"] == types.optional(TaggedValueType(list_tag_type_2, types.Int()))


MERGE_CONSTS_TEST_CASES = [
    (
        types.TypedDict(
            {
                "a": types.Const(types.Int(), 1),
            },
        ),
        types.TypedDict(
            {
                "a": types.Const(types.Int(), 2),
            },
        ),
        types.TypedDict(
            {
                "a": types.UnionType(
                    types.Const(types.Int(), 1), types.Const(types.Int(), 2)
                )
            },
        ),
    ),
    (
        types.TypedDict(
            {
                "a": types.List(
                    types.UnionType(
                        types.Const(types.Int(), 1), types.Const(types.Int(), 2)
                    )
                ),
            },
        ),
        types.TypedDict(
            {
                "a": types.List(
                    types.UnionType(
                        types.Const(types.Int(), 1), types.Const(types.Int(), 3)
                    )
                ),
            },
        ),
        types.TypedDict(
            {
                "a": types.List(
                    types.UnionType(
                        types.Const(types.Int(), 1),
                        types.Const(types.Int(), 2),
                        types.Const(types.Int(), 3),
                    )
                )
            },
        ),
    ),
    (
        types.TypedDict(
            {"a": wbmedia.ImageArtifactFileRefType({"boxLayer1": [1, 2, 3]})},
        ),
        types.TypedDict(
            {"a": wbmedia.ImageArtifactFileRefType({"boxLayer1": [1, 3, 9]})},
        ),
        types.TypedDict(
            {"a": wbmedia.ImageArtifactFileRefType({"boxLayer1": [1, 2, 3, 9]})},
        ),
    ),
]


@pytest.mark.parametrize("t1, t2, expected", MERGE_CONSTS_TEST_CASES)
def test_merge_consts(t1, t2, expected):
    assert types.merge_types(t1, t2) == expected


def test_parse_const_type():
    assert types.parse_constliteral_type({"a": [1, 2, 3]}) == types.TypedDict(
        {
            "a": types.List(
                types.union(
                    types.Const(types.Int(), 1),
                    types.Const(types.Int(), 2),
                    types.Const(types.Int(), 3),
                )
            )
        }
    )


def test_init_image():
    image_ref_type = wbmedia.ImageArtifactFileRefType({"a": [1, 2, 3]})
    assert image_ref_type.boxLayers == types.TypedDict(
        {
            "a": types.List(
                types.union(
                    types.Const(types.Int(), 1),
                    types.Const(types.Int(), 2),
                    types.Const(types.Int(), 3),
                )
            )
        }
    )

    d = image_ref_type.to_dict()
    assert d == {
        "_base_type": {"type": "Object"},
        "_is_object": True,
        "boxLayers": {"a": [1, 2, 3]},
        "boxScoreKeys": [],
        "classMap": {},
        "maskLayers": {},
        "type": "image-file",
    }


def test_deserializes_single_member_union():
    # weave0 may produce these
    assert (
        types.TypeRegistry.type_from_dict({"members": ["int"], "type": "union"})
        == types.Int()
    )


def test_wbrun_not_assignable_to_weave_run():
    from weave.legacy.weave.ops_domain import wb_domain_types

    assert not weave.types.optional(wb_domain_types.Run().WeaveType()).assign_type(
        weave.types.RunType(
            inputs=weave.types.TypedDict(property_types={}),
            history=weave.types.List(object_type=weave.types.UnknownType()),
            output=weave.types.NoneType(),
        )
    )


def test_generic_object_type():
    t = types.ObjectType(id=types.String())
    assert t.assign_type(types.RunType(output=types.Int()))
    assert not t.assign_type(types.RefType())


def test_union_auto_execute():
    assert weave.types.optional(weave.types.Timestamp()).assign_type(
        weave.types.Function(output_type=weave.types.optional(weave.types.Timestamp()))
    )


def test_load_unknown_subobj_type():
    t = weave.types.TypeRegistry.type_from_dict(
        {
            "type": "typedDict",
            "propertyTypes": {"a": "int", "b": {"type": "some_unknown_type"}},
        }
    )
    assert isinstance(t, types.TypedDict)
    assert t.property_types["a"] == types.Int()
    assert t.property_types["b"] == types.UnknownType()



"""
The purpose of this utility is to simply ensure that a W&B project exists
"""

import logging
import typing
from contextlib import contextmanager


import wandb
from wandb import errors as wandb_errors
from wandb.sdk.internal.internal_api import Api as InternalApi
from wandb.sdk.internal.internal_api import logger as wandb_logger


class AuthenticationError(wandb_errors.AuthenticationError):
    pass


class CommError(wandb_errors.CommError):
    pass


class UnableToCreateProject(Exception):
    pass


logger = logging.getLogger(__name__)


@contextmanager
def wandb_logging_disabled() -> typing.Iterator[None]:
    original_termerror = wandb.termerror
    wandb.termerror = lambda *args, **kwargs: None
    original_log_level = wandb_logger.getEffectiveLevel()
    wandb_logger.setLevel(logging.CRITICAL)
    yield None
    wandb_logger.setLevel(original_log_level)
    wandb.termerror = original_termerror


def ensure_project_exists(entity_name: str, project_name: str) -> typing.Dict[str, str]:
    with wandb_logging_disabled():
        return _ensure_project_exists(entity_name, project_name)


def _ensure_project_exists(
    entity_name: str, project_name: str
) -> typing.Dict[str, str]:
    """
    Ensures that a W&B project exists by trying to access it, returns the project_name,
    which is not guaranteed to be the same if the provided project_name contains invalid
    characters. Adheres to trace_server_interface.EnsureProjectExistsRes
    """
    wandb_logging_disabled()
    api = InternalApi({"entity": entity_name, "project": project_name})
    # Since `UpsertProject` will fail if the user does not have permission to create a project
    # we must first check if the project exists
    project = api.project(entity=entity_name, project=project_name)
    if project is None:
        exception = None
        try:
            project = api.upsert_project(entity=entity_name, project=project_name)
        except Exception as e:
            exception = e
        if project is None:
            if exception is not None:
                logger.error(f"Unable to access `{entity_name}/{project_name}`.")
                logger.error(exception.message)

                # Re-throw to not confuse the user with deep stack traces
                if isinstance(exception, wandb_errors.AuthenticationError):
                    raise AuthenticationError(exception.message)
                elif isinstance(exception, wandb_errors.CommError):
                    raise CommError(exception.message)
                else:
                    raise exception
            else:
                raise UnableToCreateProject(
                    f"Failed to create project {entity_name}/{project_name}"
                )
    return {"project_name": project["name"]}



"""Const value implementation.

You can use weave.const() to wrap a Python value. This tells the Weave type
system to encode the value as a Const type. E.g. as ConstType(StringType(), <val>) rather
than StringType()
"""

import typing

from weave.legacy.weave import weave_types as types


class Const:
    val: typing.Any

    def __init__(self, val: typing.Any):
        self.val = val


types.Const.instance_classes = Const


def const(obj: typing.Any) -> Const:
    """Mark a value as const so the Weave type system treats it as a ConstType"""
    return Const(obj)



import datetime
import typing
from math import floor

# Rules for Timestamps in Weave:
# 1. For now, we only support UTC timestamps. This is not a hard requirement and is more a
#    limitation of the current type system. We will have no problem supporting other timezones in the future
# 2. Unless otherwise specified, all timestamps are assumed to be in milliseconds.
#    * The only exception to this rule is if the timestamp is outside of the range of the
#      builtin datetime library. In this case, we will assume the timestamp is in nano or
#      micro seconds since some old sdk versions emitted this resolution
# 3. The resolution should not be finer than milliseconds. This means on disk, we store
#    timestamps as integers (for both AWL and normal lists). Furthermore, at runtime, we
#    use the datetime library which wants seconds for the constructor and produces seconds
#    for the timestamp() method. In these cases we need to be careful and do the proper division
#    or multiplication to get the correct resolution. Helper functions are provided at the bottom
#    of this file to help with this.


# all operations on datetime objects should be tz aware and in utc. If a user
# provides a non-tz aware datetime, we assume their intention is to use the
# system timezone and convert to utc.
def tz_aware_dt(dt: datetime.datetime) -> datetime.datetime:
    if dt.tzinfo is None:
        dt = dt.astimezone()  # uses system timezone
    return dt.astimezone(datetime.timezone.utc)


# These are approximately the min and max timestamps in ms for the builtin
# datetime library. we will make the very conservative assumption that any
# timestamp outside of this range, then the unit is not ms.
PY_DATETIME_MAX_MS = 50000000000000
PY_DATETIME_MIN_MS = -50000000000000


def unitless_int_to_inferred_ms(in_int: typing.Union[float, int]) -> float:
    """
    The SDK will attempt to serialize all timestamps as milliseconds. However, there were older
    versions that would sometimes produce timestamps in seconds or nanoseconds (depending on the
    datetime library used). So, in the vast majority of cases, we can assume this is a ms timestamp,
    but it is possible we are wrong. Moreover, if a user call toTimestamp on a number, in Weave0 we
    try to infer the unit of the number.

    However, in Weave1, we should use this as a learning experience and always use ms timestamps (or have
    the unit specified. If the user wants to use a different unit, they can do a multiply or divide on their
    own data, rather than having us infer it for them.)

    Previously, in Weave0, we had boundaries like:
    const timestampSecondUpperBound = 60 * 60 * 24 * 365 * 1000;
    const timestampMilliSecondUpperBound = timestampSecondUpperBound * 1000;
    const timestampMicroSecondUpperBound = timestampMilliSecondUpperBound * 1000;
    const timestampNanoSecondUpperBound = timestampMicroSecondUpperBound * 1000;

    and made the assumption that values:
        - [-inf, 31536000000] - should be interpreted as seconds
        - [31536000000, 31536000000000] - should be interpreted as milliseconds
        - [31536000000000, 31536000000000000] - should be interpreted as microseconds
        - [31536000000000000, 31536000000000000000] - should be interpreted as nanoseconds

    This has two major problems:
    1. We don't handle negative values correctly in Weave0. (notice the -inf in the first range)
    2. The seconds range introduces a nasty bug (even if we fix the negative values). Let's assume
        that we use the range [-31536000000, 31536000000] for seconds. This means that the timestamp
        range is 970/08/30 16:07:02 to 2969/05/02 17:00:00. However, that same range interpreted as
        milliseconds is 1968/12/31 16:00:00 to 1970/12/31 16:00:00. The result is that if the true unit is
        ms, then dates between those ranges would be incorrectly interpreted as seconds.


    Moving forward, since the SDK is fixed to always send ms, we should just assume that the values are in ms.
    If it is not in ms, then we should correct the data writing code or the user should do a multiply or divide.
    We can add explicit units to the types and conversion ops as well in the future.
    """
    while abs(in_int) > PY_DATETIME_MAX_MS or abs(in_int) < PY_DATETIME_MIN_MS:
        in_int = in_int / 1000
    return in_int


# While the following functions are quite simple, it is useful to have explicitly named
# functions for converting to and from ms timestamps. This makes it more clear in our
# serialization and deserialization code what exactly is happening.
def python_datetime_to_ms(dt: datetime.datetime) -> int:
    return int(dt.timestamp() * 1000)


def ms_to_python_datetime(ms: float) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(floor(ms) / 1000, tz=datetime.timezone.utc)



import weave
from weave.legacy.weave import ops_arrow
from weave.legacy.weave.wandb_interface.wandb_stream_table import StreamTable


def make_stream_table(*args, **kwargs):
    # Unit test backend does not support async logging
    return StreamTable(*args, **kwargs, _disable_async_file_stream=True)


def test_join_awls_with_stitch(user_by_api_key_in_env):
    # We use StreamTables here because they trigger projection pushdown / stitch,
    # whereas just saving an AWL does not still.
    entity_name = user_by_api_key_in_env.username
    dataset = [{"id": 1, "a": 3, "b": 4}, {"id": 2, "a": 5, "b": 6}]
    dataset_st = make_stream_table(f"{entity_name}/stream-tables/test_table_dataset")
    for row in dataset:
        dataset_st.log(row)
    dataset_st.finish()

    feedback = [{"id": 1, "a": -3, "bb": -4}, {"id": 2, "a": -5, "bb": -6}]
    feedback_st = make_stream_table(f"{entity_name}/stream-tables/test_table_feedback")
    for row in feedback:
        feedback_st.log(row)
    feedback_st.finish()

    joined = weave.legacy.weave.ops.join_2(
        dataset_st.rows(),
        feedback_st.rows(),
        lambda row: row["id"],
        lambda row: row["id"],
        "dataset",
        "feedback",
        True,
        True,
    )
    assert weave.use(joined["dataset.a"]).to_pylist_tagged() == [3, 5]
    assert weave.use(joined["dataset.aa"]).to_pylist_tagged() == [None, None]
    assert weave.use(joined["dataset.b"]).to_pylist_tagged() == [4, 6]
    assert weave.use(joined["feedback.a"]).to_pylist_tagged() == [-3, -5]
    assert weave.use(joined["feedback.bb"]).to_pylist_tagged() == [-4, -6]
    assert weave.use(joined.joinObj()).to_pylist_tagged() == [1, 2]



from PIL import Image

import weave
from weave.trace.weave_client import WeaveClient, get_ref

"""When testing types, it is important to test:
Objects:
1. Publishing Directly
2. Publishing as a property
3. Using as a cell in a table

Calls:
4. Using as inputs, output, and output component (raw)
5. Using as inputs, output, and output component (refs)

"""


def test_image_publish(client: WeaveClient) -> None:
    img = Image.new("RGB", (512, 512), "purple")
    weave.publish(img)

    ref = get_ref(img)

    assert ref is not None
    gotten_img = weave.ref(ref.uri()).get()
    assert img.tobytes() == gotten_img.tobytes()


class ImageWrapper(weave.Object):
    img: Image.Image


def test_image_as_property(client: WeaveClient) -> None:
    img = Image.new("RGB", (512, 512), "purple")
    img_wrapper = ImageWrapper(img=img)
    assert img_wrapper.img == img

    weave.publish(img_wrapper)

    ref = get_ref(img_wrapper)
    assert ref is not None

    gotten_img_wrapper = weave.ref(ref.uri()).get()
    assert gotten_img_wrapper.img.tobytes() == img.tobytes()


def test_image_as_dataset_cell(client: WeaveClient) -> None:
    img = Image.new("RGB", (512, 512), "purple")
    dataset = weave.Dataset(rows=[{"img": img}])
    assert dataset.rows[0]["img"] == img

    weave.publish(dataset)

    ref = get_ref(dataset)
    assert ref is not None

    gotten_dataset = weave.ref(ref.uri()).get()
    assert gotten_dataset.rows[0]["img"].tobytes() == img.tobytes()


@weave.op
def image_as_solo_output(publish_first: bool) -> Image.Image:
    img = Image.new("RGB", (512, 512), "purple")
    if publish_first:
        weave.publish(img)
    return img


@weave.op
def image_as_input_and_output_part(in_img: Image.Image) -> dict:
    return {"out_img": in_img}


def test_image_as_call_io(client: WeaveClient) -> None:
    non_published_img = image_as_solo_output(publish_first=False)
    img_dict = image_as_input_and_output_part(non_published_img)

    exp_bytes = non_published_img.tobytes()
    assert img_dict["out_img"].tobytes() == exp_bytes

    image_as_solo_output_call = image_as_solo_output.calls()[0]
    image_as_input_and_output_part_call = image_as_input_and_output_part.calls()[0]

    assert image_as_solo_output_call.output.tobytes() == exp_bytes
    assert image_as_input_and_output_part_call.inputs["in_img"].tobytes() == exp_bytes
    assert image_as_input_and_output_part_call.output["out_img"].tobytes() == exp_bytes


def test_image_as_call_io_refs(client: WeaveClient) -> None:
    non_published_img = image_as_solo_output(publish_first=True)
    img_dict = image_as_input_and_output_part(non_published_img)

    exp_bytes = non_published_img.tobytes()
    assert img_dict["out_img"].tobytes() == exp_bytes

    image_as_solo_output_call = image_as_solo_output.calls()[0]
    image_as_input_and_output_part_call = image_as_input_and_output_part.calls()[0]

    assert image_as_solo_output_call.output.tobytes() == exp_bytes
    assert image_as_input_and_output_part_call.inputs["in_img"].tobytes() == exp_bytes
    assert image_as_input_and_output_part_call.output["out_img"].tobytes() == exp_bytes



import numpy as np

from weave.legacy.weave import artifact_fs, artifact_wandb
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.ops_domain import table

from ...legacy.weave import types_numpy as numpy_types


def test_construct_numpy_type():
    target_type = {
        "type": "ndarray",
        "serializationPath": {
            "key": "some_key",
            "path": "a/b/c.npz",
        },
        "shape": [],
    }
    # should not raise
    t = numpy_types.NumpyArrayType.from_dict(target_type)

    # the following array should be assignable to t
    arr = np.array(1)
    t_inferred = types.TypeRegistry.type_of(arr)
    assert t_inferred.assign_type(t)


def test_construct_numpy_type_from_table_file():
    data = {
        "_type": "table",
        "column_types": {
            "params": {
                "type_map": {
                    "tca": {
                        "params": {
                            "allowed_types": [
                                {"wb_type": "none"},
                                {"wb_type": "string"},
                            ]
                        },
                        "wb_type": "union",
                    },
                    "pxe": {
                        "params": {
                            "allowed_types": [
                                {"wb_type": "none"},
                                {"wb_type": "string"},
                            ]
                        },
                        "wb_type": "union",
                    },
                    "gmi": {
                        "params": {
                            "allowed_types": [
                                {
                                    "params": {
                                        "box_layers": {
                                            "params": {"is_set": False, "val": {}},
                                            "wb_type": "const",
                                        },
                                        "box_score_keys": {
                                            "params": {"is_set": True, "val": []},
                                            "wb_type": "const",
                                        },
                                        "class_map": {
                                            "params": {"is_set": False, "val": {}},
                                            "wb_type": "const",
                                        },
                                        "mask_layers": {
                                            "params": {"is_set": False, "val": {}},
                                            "wb_type": "const",
                                        },
                                    },
                                    "wb_type": "image-file",
                                },
                                {"wb_type": "none"},
                            ]
                        },
                        "wb_type": "union",
                    },
                    "sbo": {
                        "params": {
                            "allowed_types": [
                                {"wb_type": "none"},
                                {
                                    "params": {
                                        "type_map": {
                                            "direction": {"wb_type": "number"},
                                            "image": {
                                                "params": {
                                                    "serialization_path": None,
                                                    "shape": [3, 3, 3],
                                                },
                                                "wb_type": "ndarray",
                                            },
                                            "mission": {"wb_type": "string"},
                                        }
                                    },
                                    "wb_type": "typedDict",
                                },
                            ]
                        },
                        "wb_type": "union",
                    },
                    "txet": {
                        "params": {
                            "allowed_types": [
                                {"wb_type": "none"},
                                {"wb_type": "string"},
                            ]
                        },
                        "wb_type": "union",
                    },
                }
            },
            "wb_type": "typedDict",
        },
        "columns": ["gmi", "sbo", "txet", "tca", "pxe"],
        "data": [
            [
                {
                    "_type": "image-file",
                    "format": "png",
                    "height": 192,
                    "path": "media/images/test.png",
                    "sha256": "blank",
                    "width": 352,
                },
                {
                    "direction": 0,
                    "image": [
                        [[1, 2, 3], [1, 4, 5], [1, 4, 20]],
                        [[1, 2, 3], [1, 4, 5], [10, 3, 2]],
                        [[1, 3, 3], [1, 4, 5], [1, -1, 2]],
                    ],
                    "mission": "test",
                },
                "blah",
                "blah",
                "blah2",
            ]
        ],
        "ncols": 5,
        "nrows": 1,
    }

    # generate a dummy filesystem artifact file so we can call
    # _get_rows_and_object_type_from_weave_format. that function
    # requires a filesystemartifactfile, but doesn't actually use it.
    # thus we can just use a dummy here.

    art = artifact_wandb.WandbArtifact("test")
    fs_artifact_file = artifact_fs.FilesystemArtifactFile(art, "test")
    _, object_type = table._get_rows_and_object_type_from_weave_format(
        data, fs_artifact_file
    )

    assert types.List(types.List(types.List(types.Int()))).assign_type(
        object_type.property_types["sbo"].property_types["image"]
    )
